"""
Implements a standard K-Fold cross-validator.
"""

import warnings
from copy import deepcopy
from typing import (
    Any, Dict, Generator, List, Optional, Tuple, Union
)

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.exceptions import ConvergenceWarning

from .cross_validator_interface import CrossValidator

# For type hinting sklearn-like estimators
Estimator = Any

class KFold(CrossValidator):
    """
    K-Fold cross-validator.

    This class implements the standard K-Fold cross-validation strategy, where
    the dataset is divided into `k` consecutive folds. Each fold is then
    used once as a validation set while the `k - 1` remaining folds
    form the training set.

    This implementation allows for optional shuffling.
    """

    def __init__(
        self,
        n_splits: int,
        shuffle: bool = False,
        random_seed: int = None
    ) -> None:
        """
        Initialize the K-Fold cross-validator.

        Parameters
        ----------
        n_splits : int
            Number of splits or folds for the cross-validation.
            The dataset will be divided into `n_splits` consecutive parts.
        shuffle : bool, default=False
            Whether to shuffle the data before splitting it into folds.
        random_seed : int, optional
            Seed used for random shuffling. Only used when `shuffle` is True.
        """
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_seed = random_seed
        self.rng_ = np.random.default_rng(random_seed)

    def get_n_splits(
        self,
        data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]] = None,
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> int:
        """
        Return the number of splits.

        Parameters
        ----------
        data : pd.DataFrame or dict, optional
            Dataset or dictionary of datasets.
        labels : pd.Series or dict, optional
            Labels or dictionary of labels.
        groups : np.ndarray, optional
            Group labels for the samples.

        Returns
        -------
        int
            The number of splits (`n_splits`).
        """
        return self.n_splits

    def _get_shuffled_indices(self, n_samples: int) -> np.ndarray:
        """Helper to get shuffled or sequential indices."""
        indices = np.arange(n_samples)
        if self.shuffle:
            self.rng_.shuffle(indices)
        return indices

    def _single_split(
        self,
        single_data: pd.DataFrame,
    ) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
        """
        Split a single dataset into train-test indices.

        Parameters
        ----------
        single_data : pd.DataFrame
            Input dataset.

        Yields
        -------
        Generator[Tuple[np.ndarray, np.ndarray], None, None]
            A generator where each item is a tuple of (train_indices, test_indices).
        """
        indices = self._get_shuffled_indices(single_data.shape[0])

        for test_indices in np.array_split(indices, self.n_splits):
            train_indices = np.setdiff1d(indices, test_indices)
            yield train_indices, test_indices

    def split(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> Union[
        Generator[Tuple[np.ndarray, np.ndarray], None, None],
        Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]
    ]:
        """
        Split data (or dictionary of data) into train-test indices.

        Parameters
        ----------
        data : pd.DataFrame or dict
            Dataset or dictionary of datasets.
        labels : pd.Series or dict, optional
            Labels or dictionary of labels.
        groups : np.ndarray, optional
            Group labels for the samples.

        Yields
        -------
        Generator
            - If data is a pd.DataFrame: (train_indices, test_indices)
            - If data is a dict: (key, (train_indices, test_indices))
        """
        if isinstance(data, dict):
            for key in data:
                for train_indices, test_indices in self._single_split(data[key]):
                    yield key, (train_indices, test_indices)
        else:
            for train_indices, test_indices in self._single_split(data):
                yield train_indices, test_indices

    def _single_backtest_paths(
        self,
        single_data: pd.DataFrame
    ) -> Dict[str, List[Dict[str, np.ndarray]]]:
        """
        Generate backtest paths for a single dataset.
        For K-Fold, there is only one "path".

        Parameters
        ----------
        single_data : pd.DataFrame
            Input dataset.

        Returns
        -------
        Dict[str, List[Dict[str, np.ndarray]]]
            A dictionary with one key "Path 1", containing a list of all
            (Train, Test) index dictionaries.
        """
        path_data = []
        paths = {}

        for train_indices, test_indices in self._single_split(single_data):
            path_data.append({
                "Train": train_indices,
                "Test": test_indices,
            })

        paths['Path 1'] = path_data
        return paths

    def backtest_paths(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
    ) -> Union[
        Dict[str, List[Dict[str, np.ndarray]]],
        Dict[str, Dict[str, List[Dict[str, np.ndarray]]]]
    ]:
        """
        Generate backtest paths for data or a dictionary of data.

        Parameters
        ----------
        data : pd.DataFrame or dict
            Dataset or dictionary of datasets.

        Returns
        -------
        Union[Dict, Dict[str, Dict]]
            - If data is a pd.DataFrame: A dictionary of backtest paths.
            - If data is a dict: A nested dictionary, where the first-level
              key is the dataset key, and the value is its
              dictionary of backtest paths.
        """
        if isinstance(data, dict):
            multiple_paths = {}
            for key in data:
                multiple_paths[key] = self._single_backtest_paths(data[key])
            return multiple_paths
        
        return self._single_backtest_paths(data)

    def _single_backtest_predictions(
        self,
        single_estimator: Estimator,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: Optional[np.ndarray] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Obtain backtest predictions for a single dataset.

        Parameters
        ----------
        single_estimator : Estimator
            A scikit-learn-like estimator to be trained and used for predictions.
        single_data : pd.DataFrame
            Data for the single dataset.
        single_labels : pd.Series
            Labels for the single dataset.
        single_weights : np.ndarray, optional
            Sample weights for the observations. Defaults to equal weights.
        predict_probability : bool, default=False
            If True, call `predict_proba()` instead of `predict()`.
        n_jobs : int, default=1
            The number of jobs to run in parallel.

        Returns
        -------
        Dict[str, np.ndarray]
            A dictionary with one key "Path 1", containing the contiguous
            array of out-of-sample predictions, ordered by the original
            dataset index.
        """
        if single_weights is None:
            single_weights = np.ones(len(single_data))

        def train_test_single_estimator(
            estimator_: Estimator,
            train_indices: np.ndarray,
            test_indices: np.ndarray
        ) -> Tuple[np.ndarray, np.ndarray]:
            """Train model and return (predictions, test_indices)."""
            X_train = single_data.iloc[train_indices]
            y_train = single_labels.iloc[train_indices]
            weights_train = single_weights[train_indices]

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    estimator_.fit(X_train, y_train, sample_weight=weights_train)
                except TypeError:
                    # Fallback for estimators without sample_weight
                    estimator_.fit(X_train, y_train)

            X_test = single_data.iloc[test_indices]

            if predict_probability:
                preds = estimator_.predict_proba(X_test)
            else:
                preds = estimator_.predict(X_test)

            return preds, test_indices

        # Run all folds in parallel
        results = Parallel(n_jobs=n_jobs)(
            delayed(train_test_single_estimator)(
                deepcopy(single_estimator), train_indices, test_indices
            ) for train_indices, test_indices in self._single_split(single_data)
        )

        # Unpack predictions and their corresponding original indices
        all_preds = [res[0] for res in results]
        all_indices = [res[1] for res in results]

        # Concatenate results
        path_data = np.concatenate(all_preds)
        original_indices = np.concatenate(all_indices)

        # Re-order predictions to match the original dataset order
        # This is critical if shuffle=True
        if self.shuffle:
            reorder_indices = np.argsort(original_indices)
            path_data = path_data[reorder_indices]

        paths_predictions = {'Path 1': path_data}
        return paths_predictions

    def backtest_predictions(
        self,
        estimator: Union[Estimator, Dict[str, Estimator]],
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Union[pd.Series, Dict[str, pd.Series]],
        sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]:
        """
        Generate backtest predictions for single or multiple datasets.

        Parameters
        ----------
        estimator : Estimator or dict
            Model(s) to be trained.
        data : pd.DataFrame or dict
            Input data.
        labels : pd.Series or dict
            Target labels.
        sample_weights : np.ndarray or dict, optional
            Sample weights. Defaults to None.
        predict_probability : bool, default=False
            If True, call `predict_proba()` instead of `predict()`.
        n_jobs : int, default=1
            The number of jobs to run in parallel.

        Returns
        -------
        Union[Dict, Dict[str, Dict]]
            - If data is a pd.DataFrame: A dictionary of predictions.
            - If data is a dict: A nested dictionary of predictions.
        """
        if isinstance(data, dict):
            if not isinstance(estimator, dict) or \
               not isinstance(labels, dict) or \
               (sample_weights and not isinstance(sample_weights, dict)):
                raise ValueError(
                    "If data is a dict, estimator, labels, "
                    "and sample_weights (if provided) must also be dicts."
                )
            
            multiple_paths_predictions = {}
            for key in data:
                multiple_paths_predictions[key] = self._single_backtest_predictions(
                    estimator[key],
                    data[key],
                    labels[key],
                    sample_weights[key] if sample_weights else None,
                    predict_probability,
                    n_jobs
                )
            return multiple_paths_predictions

        # Handle single dataset case
        if not isinstance(estimator, Any) or \
           not isinstance(data, pd.DataFrame) or \
           not isinstance(labels, pd.Series):
            raise ValueError(
                "If data is a DataFrame, estimator must be a single estimator "
                "and labels must be a single Series."
            )

        return self._single_backtest_predictions(
            estimator,
            data,
            labels,
            sample_weights,
            predict_probability,
            n_jobs
        )


"""
Performs backtesting on synthetic price data generated using the
Ornstein-Uhlenbeck (OU) process.
"""

from itertools import product
from random import gauss
from typing import List, Tuple

import numpy as np

def synthetic_back_testing(
    forecast: float,
    half_life: float,
    sigma: float,
    n_iteration: int = 100000,
    maximum_holding_period: int = 100,
    profit_taking_range: np.ndarray = np.linspace(0.5, 10, 20),
    stop_loss_range: np.ndarray = np.linspace(0.5, 10, 20),
    seed: int = 0,
) -> List[Tuple[float, float, float, float, float]]:
    r"""
    Perform backtesting on synthetic price data from an OU process.

    This function simulates multiple price paths based on the
    Ornstein-Uhlenbeck (OU) process and tests a grid of
    profit-taking (PT) and stop-loss (SL) levels.

    The OU process is given by:
    .. math::
        P_t = (1 - \rho) F + \rho P_{t-1} + \sigma Z_t

    Where:
    - \(P_t\) is the price at time t
    - \(F\) is the forecast price (long-term mean)
    - \(\rho\) is the autoregression coefficient
    - \(\sigma\) is the standard deviation of noise
    - \(Z_t \sim N(0, 1)\) is standard normal noise

    Parameters
    ----------
    forecast : float
        The forecasted price (long-term mean F).
    half_life : float
        The half-life of the mean-reversion process.
    sigma : float
        The standard deviation of the noise (\(\sigma\)).
    n_iteration : int, default=100000
        Number of price paths to simulate for each PT/SL combination.
    maximum_holding_period : int, default=100
        Maximum number of steps to hold the position.
    profit_taking_range : np.ndarray, default=np.linspace(0.5, 10, 20)
        An array of profit-taking levels to test.
    stop_loss_range : np.ndarray, default=np.linspace(0.5, 10, 20)
        An array of stop-loss levels to test.
    seed : int, default=0
        The initial price for all simulations.

    Returns
    -------
    List[Tuple[float, float, float, float, float]]
        A list of tuples. Each tuple contains:
        (profit_taking, stop_loss, mean_return, std_return, sharpe_ratio)
    """
    # compute ρ coefficient from half-life
    rho = 2 ** (-1.0 / half_life)
    back_test_results = []

    for profit_taking, stop_loss in product(profit_taking_range, stop_loss_range):
        stop_returns = []
        for _ in range(n_iteration):
            price, holding_period = float(seed), 0  # initial price
            while True:
                # Update price using O-U process
                price = (1 - rho) * forecast + rho * price + sigma * gauss(0, 1)
                gain = price - seed  # compute gain from initial seed price
                holding_period += 1

                # Check stop conditions
                if (
                    gain > profit_taking
                    or gain < -stop_loss
                    or holding_period > maximum_holding_period
                ):
                    stop_returns.append(gain)
                    break
        
        mean_return = np.mean(stop_returns)
        std_return = np.std(stop_returns)
        sharpe_ratio = mean_return / std_return if std_return > 0 else 0.0
        
        back_test_results.append(
            (profit_taking, stop_loss, mean_return, std_return, sharpe_ratio)
        )

    return back_test_results


"""
(DEPRECATED) Functional implementation of Information-Driven Bars.

This logic is superseded by the `ImbalanceBars` and `RunBars` classes.
Note the typo in the filename ("infomation").
"""

from typing import Tuple
import numpy as np
import pandas as pd

# Assuming these helpers are available
from RiskLabAI.utils.utilities_lopez import (
    compute_thresholds, create_ohlcv_dataframe
)

def generate_information_driven_bars(
    tick_data: pd.DataFrame,
    bar_type: str = "volume",
    initial_expected_ticks: int = 2000,
) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray]:
    """
    (DEPRECATED) Implements Information-Driven Bars.

    Parameters
    ----------
    tick_data : pd.DataFrame
        DataFrame of tick data with 'volume_labeled', 'label', 'dollar_labeled'.
    bar_type : str, default="volume"
        Type of bar: "tick", "volume", or "dollar".
    initial_expected_ticks : int, default=2000
        The initial value of expected ticks (E[T]).

    Returns
    -------
    Tuple[pd.DataFrame, np.ndarray, np.ndarray]
        - ohlcv_dataframe: The generated bars.
        - thetas_absolute: Array of absolute cumulative imbalances.
        - thresholds: Array of dynamic thresholds.
    """
    if bar_type == "volume":
        input_data = tick_data['volume_labeled']
    elif bar_type == "tick":
        input_data = tick_data['label']
    elif bar_type == "dollar":
        input_data = tick_data['dollar_labeled']
    else:
        raise ValueError("Invalid bar type. Choose 'tick', 'volume', or 'dollar'.")

    # E[b] = |mean(imbalance)|
    bar_expected_value = np.abs(input_data.mean())
    
    # Compute thresholds and grouping
    (
        times_delta,
        thetas_absolute,
        thresholds,
        times,
        thetas,
        grouping_id,
    ) = compute_thresholds(
        input_data.values, initial_expected_ticks, bar_expected_value
    )

    tick_grouped = tick_data.reset_index().assign(grouping_id=grouping_id)
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')

    ohlcv_dataframe = create_ohlcv_dataframe(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe, thetas_absolute, thresholds


"""
Generates a synthetic dataset for feature importance testing.

Based on the method from De Prado (2018).
"""

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from typing import Tuple

def get_test_dataset(
    n_features: int = 100,
    n_informative: int = 25,
    n_redundant: int = 25,
    n_samples: int = 10000,
    random_state: int = 0,
    sigma_std: float = 0.0,
) -> Tuple[pd.DataFrame, pd.Series]:
    """
    Generate a synthetic dataset with informative, redundant, and noise features.

    Parameters
    ----------
    n_features : int, default=100
        Total number of features to generate.
    n_informative : int, default=25
        Number of informative features.
    n_redundant : int, default=25
        Number of redundant features. These are created by adding
        Gaussian noise to informative features.
    n_samples : int, default=10000
        Number of samples (rows) to generate.
    random_state : int, default=0
        Random state for reproducibility.
    sigma_std : float, default=0.0
        Standard deviation of the noise added to create redundant features.

    Returns
    -------
    Tuple[pd.DataFrame, pd.Series]
        - X (pd.DataFrame): The generated feature matrix.
        - y (pd.Series): The generated labels.
    """
    rng = np.random.default_rng(random_state)

    # 1. Generate informative and noise features
    n_noise = n_features - n_informative - n_redundant
    x_array, y_array = make_classification(
        n_samples=n_samples,
        n_features=n_informative + n_noise,
        n_informative=n_informative,
        n_redundant=0,
        shuffle=False,
        random_state=random_state,
    )

    # 2. Create DataFrame
    columns = [f"I_{i}" for i in range(n_informative)]
    columns += [f"N_{i}" for i in range(n_noise)]
    x_df = pd.DataFrame(x_array, columns=columns)
    y_series = pd.Series(y_array, name="Target")

    # 3. Create redundant features
    # Randomly pick informative features to copy
    redundant_indices = rng.choice(
        range(n_informative), size=n_redundant, replace=True
    )
    
    for k, j in enumerate(redundant_indices):
        informative_col = f"I_{j}"
        noise = rng.normal(
            loc=0.0, scale=sigma_std, size=x_df.shape[0]
        )
        x_df[f"R_{k}"] = x_df[informative_col] + noise

    return x_df, y_series


DATE_TIME = "Date Time"
TIMESTAMP = 'Timestamp'
TICK_NUMBER = "Tick Number"
OPEN_PRICE = "Open"
HIGH_PRICE = "High"
LOW_PRICE = "Low"
CLOSE_PRICE = "Close"

CUMULATIVE_TICKS = "Cumulative Ticks"
CUMULATIVE_DOLLAR = "Cumulative Dollar Value"
THRESHOLD = "Threshold"

CUMULATIVE_VOLUME = "Cumulative Volume"
CUMULATIVE_BUY_VOLUME = "Cumulative Buy Volume"
CUMULATIVE_SELL_VOLUME = "Cumulative Sell Volume"

CUMULATIVE_θ = "Cumulative θ"
CUMULATIVE_BUY_θ = "Cumulative Buy θ"
CUMULATIVE_SELL_θ = "Cumulative Sell θ"

EXPECTED_IMBALANCE = "expected_imbalance"
EXPECTED_TICKS_NUMBER = "exp_num_ticks"

EXPECTED_BUY_IMBALANCE = "exp_imbalance_buy"
EXPECTED_SELL_IMBALANCE = "exp_imbalance_sell"
EXPECTED_BUY_TICKS_PROPORTION = "exp_buy_ticks_proportion"
BUY_TICKS_NUMBER = "buy_ticks_num"

N_TICKS_ON_BAR_FORMATION = "Number of ticks while bar is formed."

PREVIOUS_TICK_RULE = "Previous tick rule"
EXPECTED_IMBALANCE_WINDOW = "Expected Imbalance Window"

PREVIOUS_BARS_N_TICKS_LIST = "List of previous bars number of ticks"
PREVIOUS_TICK_IMBALANCES_LIST = "List of previous tick imbalances"

PREVIOUS_TICK_IMBALANCES_BUY_LIST = "List of previous (buy) tick imbalances"
PREVIOUS_TICK_IMBALANCES_SELL_LIST = "List of previous (sell) tick imbalances"

PREVIOUS_BARS_BUY_TICKS_PROPORTIONS_LIST = "List of previous bars buy ticks proportion"

N_PREVIOUS_BARS_FOR_EXPECTED_N_TICKS_ESTIMATION = "Window size for E[T]"




"""
Implements the Bekker-Parkinson volatility estimator, which adjusts
the Parkinson volatility by the Corwin-Schultz spread.

Reference:
    De Prado, M. (2018) Advances in Financial Machine Learning,
    page 286, snippet 19.2.
"""

from math import pi
import pandas as pd
from .corwin_schultz import beta_estimates, gamma_estimates

def sigma_estimates(beta: pd.Series, gamma: pd.Series) -> pd.Series:
    r"""
    Compute Bekker-Parkinson volatility \(\sigma\) estimates.

    .. math::
        k_2 = \sqrt{8 / \pi}
        d = 3 - 2\sqrt{2}
        \sigma = \frac{(\sqrt{2} - 1) \sqrt{\beta}}{d}
                 + \sqrt{\frac{\gamma}{k_2^2 d}}

    Parameters
    ----------
    beta : pd.Series
        \(\beta\) estimates vector from Corwin-Schultz.
    gamma : pd.Series
        \(\gamma\) estimates vector from Corwin-Schultz.

    Returns
    -------
    pd.Series
        Bekker-Parkinson volatility \(\sigma\) estimates.
    """
    k2 = (8 / pi) ** 0.5
    denominator = 3 - 2 * (2**0.5)

    term1 = (2**0.5 - 1) * (beta**0.5) / denominator
    term2 = (gamma / (k2**2 * denominator)) ** 0.5
    
    sigma = term1 + term2
    sigma[sigma < 0] = 0.0 # Floor at zero

    return sigma


def bekker_parkinson_volatility_estimates(
    high_prices: pd.Series, low_prices: pd.Series, window_span: int = 20
) -> pd.Series:
    """
    Compute Bekker-Parkinson volatility estimates from high and low prices.

    This function first calculates the Corwin-Schultz \(\beta\) and \(\gamma\)
    parameters and then uses them to compute the volatility estimates.

    Parameters
    ----------
    high_prices : pd.Series
        Time series of high prices.
    low_prices : pd.Series
        Time series of low prices.
    window_span : int, default=20
        Rolling window span for \(\beta\) estimation.

    Returns
    -------
    pd.Series
        Bekker-Parkinson volatility estimates.
    """
    beta = beta_estimates(high_prices, low_prices, window_span)
    gamma = gamma_estimates(high_prices, low_prices)

    return sigma_estimates(beta, gamma)


"""
Provides a simple terminal-style progress bar.
"""

import sys
import time
from typing import Optional

def progress_bar(
    current_progress: int,
    total_progress: int,
    start_time: float,
    bar_length: int = 20,
) -> None:
    r"""
    Display a terminal-style progress bar.

    Shows completion percentage and estimated time remaining.

    Parameters
    ----------
    current_progress : int
        Current iteration number (e.g., `i`).
    total_progress : int
        Total number of items to process.
    start_time : float
        The time the task started (from `time.time()`).
    bar_length : int, default=20
        Length of the progress bar in characters.
    """
    if total_progress == 0:
        return
        
    percentage = current_progress / total_progress
    
    # Handle the very first iteration
    if percentage == 0:
        arrow = ""
    else:
        arrow = "-" * int(round(percentage * bar_length) - 1) + ">"
    
    spaces = " " * (bar_length - len(arrow))
    elapsed_time_sec = time.time() - start_time

    # Calculate remaining time
    if current_progress == 0:
        remaining_time_str = "Calculating..."
    else:
        remaining_time_min = (
            (elapsed_time_sec / current_progress)
            * (total_progress - current_progress)
        ) / 60
        remaining_time_str = f"{remaining_time_min:.0f} minutes remaining"

    # Assemble and print the bar
    sys.stdout.write(
        f"\rCompleted: [{arrow + spaces}] {percentage*100:.0f}% - {remaining_time_str}."
    )
    
    if current_progress == total_progress:
        sys.stdout.write(
            f"\rCompleted: [{'-' * bar_length}] 100% - Task completed!          \n"
        )
    
    sys.stdout.flush()


"""
RiskLabAI Data Structures Module

Implements various bar types (Standard, Time, Imbalance, Run) and
related financial data structures and algorithms from de Prado (2018).

Note: This module contains both a modern, class-based implementation
(e.g., StandardBars, ImbalanceBars) and an older, functional implementation
(e.g., generate_tick_bar_dataframe, infomation_driven_bars.py).
The class-based approach is recommended.
"""

from .abstract_bars import AbstractBars
from .abstract_information_driven_bars import AbstractInformationDrivenBars
from .abstract_imbalance_bars import AbstractImbalanceBars
from .abstract_run_bars import AbstractRunBars

# Concrete Class Implementations
from .standard_bars import StandardBars
from .time_bars import TimeBars
from .imbalance_bars import ExpectedImbalanceBars, FixedImbalanceBars
from .run_bars import ExpectedRunBars, FixedRunBars

# Hedging
from .hedging import pca_weights

# Filtering (Duplicated)
from .filtering_lopez import symmetric_cusum_filter

# Utilities & Deprecated Functional Bars
from .utilities_lopez import (
    compute_thresholds,
    create_ohlcv_dataframe
)
from .data_structures_lopez import (
    progress_bar,
    ewma,
    compute_grouping,
    generate_information_driven_bars as functional_info_bars, # Renamed
    generate_time_bar, # Duplicated
    generate_tick_bar, # Duplicated
    generate_volume_bar, # Duplicated
    generate_dollar_bar, # Duplicated
    calculate_pca_weights, # Duplicated
    events, # Duplicated
)
from .standard_bars_lopez import (
    generate_dollar_bar_dataframe,
    generate_tick_bar_dataframe,
    generate_time_bar_dataframe,
    generate_volume_bar_dataframe,
)
# Note: Filename has a typo
from .infomation_driven_bars import generate_information_driven_bars


__all__ = [
    # Abstract Classes
    "AbstractBars",
    "AbstractInformationDrivenBars",
    "AbstractImbalanceBars",
    "AbstractRunBars",
    
    # Concrete Bar Classes
    "StandardBars",
    "TimeBars",
    "ExpectedImbalanceBars",
    "FixedImbalanceBars",
    "ExpectedRunBars",
    "FixedRunBars",
    
    # Hedging
    "pca_weights",
    
    # Filtering
    "symmetric_cusum_filter",
    
    # Utilities
    "create_ohlcv_dataframe",
    "ewma",
    "progress_bar",

    # Deprecated/Duplicated Functional API
    "compute_thresholds",
    "compute_grouping",
    "generate_information_driven_bars",
    "functional_info_bars",
    "generate_time_bar",
    "generate_tick_bar",
    "generate_volume_bar",
    "generate_dollar_bar",
    "generate_dollar_bar_dataframe",
    "generate_tick_bar_dataframe",
    "generate_time_bar_dataframe",
    "generate_volume_bar_dataframe",
    "calculate_pca_weights",
    "events",
]


"""
Implements the Corwin and Schultz (2012) bid-ask spread estimator.

Reference:
    Corwin, S. A., & Schultz, P. (2012). A simple way to estimate
    bid-ask spreads from daily high and low prices.
    The Journal of Finance, 67(2), 719-760.
"""

import numpy as np
import pandas as pd

def beta_estimates(
    high_prices: pd.Series, low_prices: pd.Series, window_span: int
) -> pd.Series:
    r"""
    Estimate \(\beta\) (sum of squared high-low log-ratios).

    .. math::
        \beta_t = \sum_{j=t-1}^{t} [\ln(H_j / L_j)]^2

    This is then averaged over the `window_span`.

    Parameters
    ----------
    high_prices : pd.Series
        Time series of high prices.
    low_prices : pd.Series
        Time series of low prices.
    window_span : int
        Rolling window span for averaging.

    Returns
    -------
    pd.Series
        The estimated \(\beta\) vector.
    """
    log_ratios_sq = np.log(high_prices / low_prices) ** 2
    
    # Sum of current and previous day's squared log-ratio
    beta = log_ratios_sq.rolling(window=2).sum()
    
    # Average over the window span
    beta = beta.rolling(window=window_span).mean()
    return beta


def gamma_estimates(high_prices: pd.Series, low_prices: pd.Series) -> pd.Series:
    r"""
    Estimate \(\gamma\) (squared log-ratio of two-day high/low).

    .. math::
        \gamma_t = [\ln(\max(H_t, H_{t-1}) / \min(L_t, L_{t-1}))]^2

    Parameters
    ----------
    high_prices : pd.Series
        Time series of high prices.
    low_prices : pd.Series
        Time series of low prices.

    Returns
    -------
    pd.Series
        The estimated \(\gamma\) vector.
    """
    high_prices_max = high_prices.rolling(window=2).max()
    low_prices_min = low_prices.rolling(window=2).min()
    gamma = np.log(high_prices_max / low_prices_min) ** 2
    return gamma


def alpha_estimates(beta: pd.Series, gamma: pd.Series) -> pd.Series:
    r"""
    Estimate \(\alpha\) from \(\beta\) and \(\gamma\).

    .. math::
        \alpha = \frac{\sqrt{2\beta} - \sqrt{\beta}}{3 - 2\sqrt{2}}
                 - \sqrt{\frac{\gamma}{3 - 2\sqrt{2}}}

    Parameters
    ----------
    beta : pd.Series
        \(\beta\) estimates vector.
    gamma : pd.Series
        \(\gamma\) estimates vector.

    Returns
    -------
    pd.Series
        The estimated \(\alpha\) vector, floored at 0.
    """
    denominator = 3 - 2 * (2**0.5)
    term1 = ((2**0.s5) - 1) * (beta**0.5) / denominator
    term2 = (gamma / denominator) ** 0.5
    
    alpha = term1 - term2
    alpha[alpha < 0] = 0.0  # Floor at zero
    return alpha


def corwin_schultz_estimator(
    high_prices: pd.Series, low_prices: pd.Series, window_span: int = 20
) -> pd.Series:
    r"""
    Estimate the bid-ask spread using the Corwin and Schultz (2012) method.

    .. math::
        S = \frac{2(e^\alpha - 1)}{1 + e^\alpha}

    Parameters
    ----------
    high_prices : pd.Series
        Time series of high prices.
    low_prices : pd.Series
        Time series of low prices.
    window_span : int, default=20
        Rolling window span for \(\beta\) estimation.

    Returns
    -------
    pd.Series
        The estimated spread vector.
    """
    beta = beta_estimates(high_prices, low_prices, window_span)
    gamma = gamma_estimates(high_prices, low_prices)
    alpha = alpha_estimates(beta, gamma)
    
    # Calculate spread
    spread = 2 * (np.exp(alpha) - 1) / (1 + np.exp(alpha))
    return spread


import time
import numpy as np
import pandas as pd
from math import ceil
from scipy import stats as ss
from sklearn.linear_model import LogisticRegression
from tqdm import tqdm
import ta
import itertools
import warnings
from typing import Dict, Union, Tuple, List, Any, Optional
from scipy.stats import kendalltau
from memory_profiler import memory_usage
import subprocess

from RiskLabAI.data.differentiation import fractionally_differentiated_log_price
from RiskLabAI.data.labeling import daily_volatility_with_log_returns, cusum_filter_events_dynamic_threshold, vertical_barrier, meta_events, meta_labeling
from RiskLabAI.data.weights import sample_weight_absolute_return_meta_labeling
from RiskLabAI.utils import determine_strategy_side
from RiskLabAI.backtest.validation import CrossValidatorController
from RiskLabAI.backtest import probability_of_backtest_overfitting, probabilistic_sharpe_ratio, benchmark_sharpe_ratio, sharpe_ratio, strategy_bet_sizing

def financial_features_backtest_overfitting_simulation(
    prices: pd.Series, 
    noise_scale: float = 0.0, 
    random_state: Optional[int] = None
) -> pd.DataFrame:
    """
    Create a DataFrame of financial features from the given price series.

    Args:
        prices (pd.Series): Time series of asset prices.
        noise_scale (float): Scale of Gaussian noise to be added to the features. Default is 0.0.
        random_state (Optional[int]): Seed for random number generator. Default is None.

    Returns:
        pd.DataFrame: DataFrame containing the computed financial features.
    """
    rng = np.random.default_rng(random_state)
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        # Features
        features = pd.DataFrame()
        features['FracDiff'] = fractionally_differentiated_log_price(prices)
        features['Volatility'] = daily_volatility_with_log_returns(prices, 100)
        features['Z-Score'] = (prices - prices.rolling(20).mean()) / prices.rolling(20).std()
        macd_line = np.log(prices.ewm(span=12).mean() / prices.ewm(span=26).mean())
        signal_line = macd_line.ewm(span=9).mean()
        features["Log MACD Histogram"] = macd_line - signal_line
        features["ADX"] = ta.trend.ADXIndicator(prices, prices, prices, fillna=True).adx()
        features["RSI"] = ta.momentum.RSIIndicator(prices, fillna=True).rsi()
        features["CCI"] = ta.trend.CCIIndicator(prices, prices, prices, fillna=True).cci()
        stochastic = ta.momentum.StochasticOscillator(prices, prices, prices, fillna=True)
        features["Stochastic"] = stochastic.stoch()
        features["ROC"] = ta.momentum.ROCIndicator(prices, fillna=True).roc()
        features["ATR"] = ta.volatility.AverageTrueRange(prices, prices, prices, fillna=True).average_true_range()
        features["Log DPO"] = np.log(prices.rolling(11).mean() / prices.rolling(20).mean())
        
        # 1. MACD Crossovers:
        features["MACD Position"] = 0  # default to no crossover
        features.loc[features["Log MACD Histogram"] >= 0, "MACD Position"] = 1
        features.loc[features["Log MACD Histogram"] < 0, "MACD Position"] = -1

        # 2. ADX Trend Strength:
        features["ADX Strength"] = 0
        features.loc[features["ADX"] > 25, "ADX Strength"] = 1
        features.loc[features["ADX"] < 25, "ADX Strength"] = -1

        # 3. RSI Overbought/Oversold:
        features["RSI Signal"] = 0
        features.loc[features["RSI"] > 70, "RSI Signal"] = 1
        features.loc[features["RSI"] < 30, "RSI Signal"] = -1

        # 4. CCI Overbought/Oversold:
        features["CCI Signal"] = 0
        features.loc[features["CCI"] > 100, "CCI Signal"] = 1
        features.loc[features["CCI"] < -100, "CCI Signal"] = -1

        # 5. Stochastic Oscillator Overbought/Oversold:
        stochastic_signal = stochastic.stoch_signal()
        features["Stochastic Signal"] = 0
        features.loc[stochastic_signal > 80, "Stochastic Signal"] = 1
        features.loc[stochastic_signal < 20, "Stochastic Signal"] = -1

        # 6. ROC Momentum Shifts:
        features["ROC Momentum"] = 0
        features.loc[features["ROC"] > 0, "ROC Momentum"] = 1  # Positive momentum
        features.loc[features["ROC"] < 0, "ROC Momentum"] = -1  # Negative momentum
        ichimoku = ta.trend.IchimokuIndicator(prices, prices, visual=False, fillna=True)

        # 1. Kumo Breakouts
        features["Kumo Breakout"] = 0
        senkou_span_a = ichimoku.ichimoku_a()
        senkou_span_b = ichimoku.ichimoku_b()
        features.loc[(prices > senkou_span_a) & (prices > senkou_span_b), "Kumo Breakout"] = 1
        features.loc[(prices < senkou_span_a) & (prices < senkou_span_b), "Kumo Breakout"] = -1

        # 2. TK Crosses
        features["TK Position"] = 0
        tenkan_sen = ichimoku.ichimoku_conversion_line()
        kijun_sen = ichimoku.ichimoku_base_line()
        features.loc[tenkan_sen >= kijun_sen, "TK Position"] = 1
        features.loc[tenkan_sen < kijun_sen, "TK Position"] = -1

        # 4. Price Relative to Kumo
        features["Price Kumo Position"] = 0
        features.loc[(prices > senkou_span_a) & (prices > senkou_span_b), "Price Kumo Position"] = 1
        features.loc[(prices < senkou_span_a) & (prices < senkou_span_b), "Price Kumo Position"] = -1

        # 5. Cloud Thickness
        features["Cloud Thickness"] = np.log(senkou_span_a / senkou_span_b)

        # 6. Momentum Confirmation
        features["Momentum Confirmation"] = 0
        features.loc[(tenkan_sen > senkou_span_a) & (prices > senkou_span_a), "Momentum Confirmation"] = 1
        features.loc[(tenkan_sen < senkou_span_a) & (prices < senkou_span_a), "Momentum Confirmation"] = -1

    if noise_scale > 0.0:
        for col in features.columns:
            noise = rng.normal(loc=0, scale=noise_scale * features[col].std(), size=features[col].shape)
            features[col] += noise
    
    return features

def backtest_overfitting_simulation_results(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    cross_validators: Dict[str, Any],
    noise_scale: float = 0.0,
    random_state: int = None,
    n_jobs: int = 1
) -> Dict[str, List[Dict[str, Any]]]:
    """
    Conducts a simulation to evaluate the performance of trading strategies and models.

    This function simulates a trading environment to assess various cross-validation methods in the context of financial analytics. It uses a set of market regime parameters and machine learning models to backtest trading strategies and compute metrics indicative of overfitting.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior, including window sizes and flags for mean reversion.
        models (dict): A collection of machine learning models and their associated parameters.
        cross_validators (dict): A dictionary of cross-validation methods.
        noise_scale (float): Scale of Gaussian noise to be added to the features.
        random_state (int): Seed for random number generator.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Dict[str, List[Dict[str, Any]]]: A dictionary containing the results of the backtest for each cross-validation method tested.
    """

    volatility = daily_volatility_with_log_returns(prices, 100)
    filter_threshold = 1.8
    moelcules = cusum_filter_events_dynamic_threshold(np.log(prices), filter_threshold * volatility)
    vertical_barriers = vertical_barrier(prices, moelcules, 20)
    features = financial_features_backtest_overfitting_simulation(prices, noise_scale=noise_scale, random_state=random_state)

    results = {cv: [] for cv in cross_validators.keys()}

    # Iterate over each strategy parameter combination
    strategy_parameters_keys, strategy_parameters_values = zip(*strategy_parameters.items())
    for strategy_parameters_value in itertools.product(*strategy_parameters_values):
        strategy_params = dict(zip(strategy_parameters_keys, strategy_parameters_value))
        if (strategy_params['fast_window'] == strategy_parameters['fast_window'][0] and strategy_params['slow_window'] == strategy_parameters['slow_window'][0]) or \
        (strategy_params['fast_window'] == strategy_parameters['fast_window'][1] and strategy_params['slow_window'] == strategy_parameters['slow_window'][1]) or \
        (strategy_params['fast_window'] == strategy_parameters['fast_window'][2] and strategy_params['slow_window'] == strategy_parameters['slow_window'][2]) or \
        (strategy_params['fast_window'] == strategy_parameters['fast_window'][3] and strategy_params['slow_window'] == strategy_parameters['slow_window'][3]):
            
            strategy_sides = determine_strategy_side(prices, **strategy_params)
        else:
            continue    

        triple_barrier_events = meta_events(prices, moelcules, [0.5, 1.5], volatility, 0, 1, vertical_barriers, strategy_sides)
        labels = meta_labeling(triple_barrier_events, prices)
        sample_weights = sample_weight_absolute_return_meta_labeling(triple_barrier_events['End Time'], prices, moelcules)

        index = features.loc[moelcules].dropna().index.intersection(labels.dropna().index).intersection(sample_weights.dropna().index)
        data = features.loc[index]
        target = labels.loc[index]['Label']
        weights = sample_weights.loc[index]
        times = labels.loc[index]['End Time']
        external_feature = volatility.loc[index]

        # Iterate over each model and hyperparameter configuration
        for model_name, model_details in models.items():
            model = model_details['Model']
            param_grid = model_details['Parameters']

            # Generate all combinations of hyperparameters
            model_keys, model_values = zip(*param_grid.items())
            for model_value in itertools.product(*model_values):
                params = dict(zip(model_keys, model_value))
                model.set_params(**params)

                for cross_validator_type, cross_validator in cross_validators.items():
                    if 'times' in cross_validator.__dict__:
                        cross_validator.times = times
                    if 'external_feature' in cross_validator.__dict__:
                        cross_validator.external_feature = external_feature    
                    predictions = cross_validator.backtest_predictions(model, data, target, weights, predict_probability=True, n_jobs=n_jobs)
                    probabilities = pd.Series(np.vstack(list(map(lambda x: x[:, 1], predictions.values()))).mean(axis=0), times.index).dropna()
                    positions = strategy_bet_sizing(prices.index, times.loc[probabilities.index], strategy_sides[probabilities.index], probabilities)
                    strategy_log_returns = (np.log(prices).diff() * positions.shift()).dropna()
                    results[cross_validator_type].append({
                        'Trial Info.' : {
                            'Strategy Parameters' : strategy_params,
                            'Model Name': model_name,
                            'Model Parameters': params,
                        },
                        'Returns': strategy_log_returns
                    })

    return results

def overall_backtest_overfitting_simulation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    noise_scale: float = 0.0,
    random_state: int = None,
    n_jobs: int = 1
) -> Tuple[Dict[str, float], Dict[str, float]]:
    """
    Conducts an overall backtest overfitting simulation to calculate the metrics.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        noise_scale (float): Scale of Gaussian noise to be added to the features.
        random_state (int): Seed for random number generator.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[Dict[str, float], Dict[str, float]]: A tuple containing two dictionaries, one for the Probability of Backtest Overfitting (PBO) and the other for the Deflated Sharpe Ratio (DSR), for each cross-validation method tested.
    """

    cross_validators = {
        'Walk-Forward' : CrossValidatorController(
            'walkforward',
            n_splits=4,
        ).cross_validator,
        'K-Fold' : CrossValidatorController(
            'kfold',
            n_splits=4,
        ).cross_validator,
        'Purged K-Fold' : CrossValidatorController(
            'purgedkfold',
            n_splits=4,
            times=None,
            embargo=0.02
        ).cross_validator,
        'Combinatorial Purged' : CrossValidatorController(
            'combinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02
        ).cross_validator,
    }

    results = backtest_overfitting_simulation_results(prices, strategy_parameters, models, cross_validators, noise_scale, random_state, n_jobs=n_jobs)


    cv_deflated_sr = {}
    cv_pbo = {}

    for cv, trials in results.items():
        performances = pd.concat([trial['Returns'] for trial in trials], axis=1)   

        sharpe_ratios = performances.apply(lambda y: sharpe_ratio(y.values))    
        benchmark_sr = benchmark_sharpe_ratio(sharpe_ratios)
        best_strategy_index = sharpe_ratios.idxmax()
        cv_deflated_sr[cv] = probabilistic_sharpe_ratio(
            sharpe_ratios.loc[best_strategy_index], 
            benchmark_sr, len(performances), 
            ss.skew(performances[best_strategy_index]), 
            ss.kurtosis(performances[best_strategy_index]),
            return_test_statistic=True
        )

        pbo, logit_values = probability_of_backtest_overfitting(performances.values, risk_free_return=step_risk_free_rate)    
        cv_pbo[cv] = pbo

    return cv_pbo, cv_deflated_sr    

def temporal_backtest_overfitting_simulation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    overfitting_partitions_length: int,
    n_jobs: int = 1
) -> Tuple[Dict[str, List[float]], Dict[str, List[float]]]:
    """
    Conducts a temporal backtest overfitting simulation to calculate the metrics in chunks.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        overfitting_partitions_length (int): The number of partitions to divide the dataset into for temporal overfitting analysis.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[Dict[str, List[float]], Dict[str, List[float]]]: A tuple containing two dictionaries, one for the Probability of Backtest Overfitting (PBO) and the other for the Deflated Sharpe Ratio (DSR), for each cross-validation method tested.
    """
    cross_validators = {
        'Walk-Forward' : CrossValidatorController(
            'walkforward',
            n_splits=4,
        ).cross_validator,
        'K-Fold' : CrossValidatorController(
            'kfold',
            n_splits=4,
        ).cross_validator,
        'Purged K-Fold' : CrossValidatorController(
            'purgedkfold',
            n_splits=4,
            times=None,
            embargo=0.02
        ).cross_validator,
        'Combinatorial Purged' : CrossValidatorController(
            'combinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02
        ).cross_validator,
    }

    results = backtest_overfitting_simulation_results(prices, strategy_parameters, models, cross_validators, n_jobs=n_jobs)

    cv_deflated_sr = {cv: [] for cv in results.keys()}
    cv_pbo = {cv: [] for cv in results.keys()}

    for cv, trials in results.items():
        performances = pd.concat([trial['Returns'] for trial in trials], axis=1)
        
        # Calculate the number of chunks using ceil
        n_chunks = ceil(performances.shape[0] / overfitting_partitions_length)
        for chunk in np.array_split(performances, n_chunks):
            sharpe_ratios = chunk.apply(lambda y: sharpe_ratio(y.values))
            benchmark_sr = benchmark_sharpe_ratio(sharpe_ratios)
            best_strategy_index = sharpe_ratios.idxmax()
            
            deflated_sr = probabilistic_sharpe_ratio(
                sharpe_ratios.loc[best_strategy_index], 
                benchmark_sr, len(chunk), 
                ss.skew(chunk[best_strategy_index]), 
                ss.kurtosis(chunk[best_strategy_index]),
                return_test_statistic=True
            )
            cv_deflated_sr[cv].append(deflated_sr)
            
            pbo, logit_values = probability_of_backtest_overfitting(chunk.values, risk_free_return=step_risk_free_rate, n_jobs=n_jobs)
            cv_pbo[cv].append(pbo)

    return cv_pbo, cv_deflated_sr

def time_temporal_backtest_overfitting_simulation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    overfitting_partitions_duration: str = 'A',  # Annual grouping by default
    n_jobs: int = 1
) -> Tuple[Dict[str, pd.Series], Dict[str, pd.Series]]:
    """
    Conducts a time-temporal backtest overfitting simulation to calculate the metrics in time-indexed chunks.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        overfitting_partitions_duration (str): The frequency for time-based grouping to divide the dataset into for temporal overfitting analysis.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[Dict[str, pd.Series], Dict[str, pd.Series]]: A tuple containing two dictionaries, one for the Probability of Backtest Overfitting (PBO) and the other for the Deflated Sharpe Ratio (DSR), for each cross-validation method tested, indexed by time.
    """
    cross_validators = {
        'Walk-Forward' : CrossValidatorController(
            'walkforward',
            n_splits=4,
        ).cross_validator,
        'K-Fold' : CrossValidatorController(
            'kfold',
            n_splits=4,
        ).cross_validator,
        'Purged K-Fold' : CrossValidatorController(
            'purgedkfold',
            n_splits=4,
            times=None,
            embargo=0.02
        ).cross_validator,
        'Combinatorial Purged' : CrossValidatorController(
            'combinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02
        ).cross_validator,
    }

    results = backtest_overfitting_simulation_results(prices, strategy_parameters, models, cross_validators, n_jobs=n_jobs)

    cv_deflated_sr = {cv: pd.Series(dtype=float) for cv in results.keys()}
    cv_pbo = {cv: pd.Series(dtype=float) for cv in results.keys()}

    for cv, trials in results.items():
        performances = pd.concat([trial['Returns'] for trial in trials], axis=1)

        # Group by the specified duration and calculate metrics
        grouped = performances.groupby(pd.Grouper(freq=overfitting_partitions_duration))
        for timestamp, chunk in grouped:
            sharpe_ratios = chunk.apply(lambda y: sharpe_ratio(y.values))
            benchmark_sr = benchmark_sharpe_ratio(sharpe_ratios)
            best_strategy_index = sharpe_ratios.idxmax()

            deflated_sr = probabilistic_sharpe_ratio(
                sharpe_ratios.loc[best_strategy_index],
                benchmark_sr, len(chunk),
                ss.skew(chunk[best_strategy_index]),
                ss.kurtosis(chunk[best_strategy_index]),
                return_test_statistic=True
            )
            cv_deflated_sr[cv].at[timestamp] = deflated_sr

            pbo, logit_values = probability_of_backtest_overfitting(chunk.values, risk_free_return=step_risk_free_rate, n_jobs=1)
            cv_pbo[cv].at[timestamp] = pbo

    return cv_pbo, cv_deflated_sr

def varying_embargo_backtest_overfitting_simulation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    embargo_values: List[float],
    n_jobs: int = 1
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Conducts a backtest overfitting simulation with varying embargo values to calculate the metrics.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        embargo_values (List[float]): List of embargo values to test.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: Two DataFrames containing PBO and DSR values for each embargo value and cross-validation method.
    """

    volatility = daily_volatility_with_log_returns(prices, 100)
    filter_threshold = 1.8
    moelcules = cusum_filter_events_dynamic_threshold(np.log(prices), filter_threshold * volatility)
    vertical_barriers = vertical_barrier(prices, moelcules, 20)
    features = financial_features_backtest_overfitting_simulation(prices)

    cv_pbo_embargo = pd.DataFrame(index=embargo_values, columns=['Purged K-Fold', 'Combinatorial Purged'])
    cv_deflated_sr_embargo = pd.DataFrame(index=embargo_values, columns=['Purged K-Fold', 'Combinatorial Purged'])

    for embargo in embargo_values:
        results = {
            'Purged K-Fold' : [],
            'Combinatorial Purged' : [],
        }

        cross_validators = {
            'Purged K-Fold' : CrossValidatorController(
                'purgedkfold',
                n_splits=4,
                times=None,
                embargo=embargo
            ).cross_validator,
            'Combinatorial Purged' : CrossValidatorController(
                'combinatorialpurged',
                n_splits=8,
                n_test_groups=2,
                times=None,
                embargo=embargo
            ).cross_validator,
        }

        # Iterate over each strategy parameter combination
        strategy_parameters_keys, strategy_parameters_values = zip(*strategy_parameters.items())
        for strategy_parameters_value in itertools.product(*strategy_parameters_values):
            strategy_params = dict(zip(strategy_parameters_keys, strategy_parameters_value))
            if (strategy_params['fast_window'] == strategy_parameters['fast_window'][0] and strategy_params['slow_window'] == strategy_parameters['slow_window'][0]) or \
            (strategy_params['fast_window'] == strategy_parameters['fast_window'][1] and strategy_params['slow_window'] == strategy_parameters['slow_window'][1]) or \
            (strategy_params['fast_window'] == strategy_parameters['fast_window'][2] and strategy_params['slow_window'] == strategy_parameters['slow_window'][2]) or \
            (strategy_params['fast_window'] == strategy_parameters['fast_window'][3] and strategy_params['slow_window'] == strategy_parameters['slow_window'][3]):
                
                strategy_sides = determine_strategy_side(prices, **strategy_params)
            else:
                continue    

            triple_barrier_events = meta_events(prices, moelcules, [0.5, 1.5], volatility, 0, 1, vertical_barriers, strategy_sides)
            labels = meta_labeling(triple_barrier_events, prices)
            sample_weights = sample_weight_absolute_return_meta_labeling(triple_barrier_events['End Time'], prices, moelcules)

            index = features.loc[moelcules].dropna().index.intersection(labels.dropna().index).intersection(sample_weights.dropna().index)
            data = features.loc[index]
            target = labels.loc[index]['Label']
            weights = sample_weights.loc[index]
            times = labels.loc[index]['End Time']

            # Iterate over each model and hyperparameter configuration
            for model_name, model_details in models.items():
                model = model_details['Model']
                param_grid = model_details['Parameters']

                # Generate all combinations of hyperparameters
                model_keys, model_values = zip(*param_grid.items())
                for model_value in itertools.product(*model_values):
                    params = dict(zip(model_keys, model_value))
                    model.set_params(**params)

                    for cross_validator_type, cross_validator in cross_validators.items():
                        if 'times' in cross_validator.__dict__:
                            cross_validator.times = times
                        predictions = cross_validator.backtest_predictions(model, data, target, weights, predict_probability=True, n_jobs=n_jobs)
                        probabilities = pd.Series(np.vstack(list(map(lambda x: x[:, 1], predictions.values()))).mean(axis=0), times.index).dropna()
                        positions = strategy_bet_sizing(prices.index, times.loc[probabilities.index], strategy_sides[probabilities.index], probabilities)
                        strategy_log_returns = (np.log(prices).diff() * positions.shift()).dropna()
                        results[cross_validator_type].append({
                            'Trial Info.' : {
                                'Strategy Parameters' : strategy_params,
                                'Model Name': model_name,
                                'Model Parameters': params,
                            },
                            'Returns': strategy_log_returns
                        })

        cv_deflated_sr = {}
        cv_pbo = {}

        for cv, trials in results.items():
            performances = pd.concat([trial['Returns'] for trial in trials], axis=1)   

            sharpe_ratios = performances.apply(lambda y: sharpe_ratio(y.values))    
            benchmark_sr = benchmark_sharpe_ratio(sharpe_ratios)
            best_strategy_index = sharpe_ratios.idxmax()
            cv_deflated_sr[cv] = probabilistic_sharpe_ratio(
                sharpe_ratios.loc[best_strategy_index], 
                benchmark_sr, len(performances), 
                ss.skew(performances[best_strategy_index]), 
                ss.kurtosis(performances[best_strategy_index]),
                return_test_statistic=True
            )

            pbo, logit_values = probability_of_backtest_overfitting(performances.values, risk_free_return=step_risk_free_rate, n_jobs=1)    
            cv_pbo[cv] = pbo

        for cv in results.keys():
            cv_pbo_embargo.loc[embargo, cv] = cv_pbo[cv]
            cv_deflated_sr_embargo.loc[embargo, cv] = cv_deflated_sr[cv]

    return cv_pbo_embargo, cv_deflated_sr_embargo

def sharpe_ratio(returns, risk_free_rate=0):
    """Calculate the Sharpe ratio of the given returns."""
    return (returns.mean() - risk_free_rate) / returns.std()

def sortino_ratio(returns, risk_free_rate=0):
    """Calculate the Sortino ratio of the given returns."""
    downside_returns = returns[returns < 0]
    expected_return = returns.mean() - risk_free_rate
    downside_risk = np.sqrt((downside_returns ** 2).mean())
    return expected_return / downside_risk

def expected_shortfall(returns, step_risk_free_rate, confidence_level=0.05):
    """Calculate the expected shortfall (conditional VaR) of the given returns."""
    var = np.percentile(returns, 100 * confidence_level)
    es = returns[returns <= var].mean()
    return es

def backtest_overfitting_simulation_financial_metrics_rank_correlation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    n_jobs: int = 1
) -> pd.DataFrame:
    """
    Conducts a backtest overfitting simulation and calculates the rank correlation of financial metrics.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        pd.DataFrame: DataFrame containing the rank correlations for each cross-validation method and each metric.
    """
    # Run the backtest overfitting simulation to get results
    cross_validators = {
        'Walk-Forward' : CrossValidatorController(
            'walkforward',
            n_splits=4,
        ).cross_validator,
        'K-Fold' : CrossValidatorController(
            'kfold',
            n_splits=4,
        ).cross_validator,
        'Purged K-Fold' : CrossValidatorController(
            'purgedkfold',
            n_splits=4,
            times=None,
            embargo=0.02
        ).cross_validator,
        'Combinatorial Purged' : CrossValidatorController(
            'combinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02
        ).cross_validator,
    }

    results = backtest_overfitting_simulation_results(prices, strategy_parameters, models, cross_validators, n_jobs=n_jobs)
    
    metrics = {
        'Sharpe Ratio': sharpe_ratio,
        'Sortino Ratio': sortino_ratio,
        'Expected Shortfall': expected_shortfall
    }

    rank_correlations = {cv: {} for cv in results.keys()}

    for cv, trials in results.items():
        performances = pd.concat([trial['Returns'] for trial in trials], axis=1)

        # Split the data into two halves
        midpoint = len(performances) // 2
        first_half = performances.iloc[:midpoint]
        second_half = performances.iloc[midpoint:]

        # Calculate the metrics for each half
        first_half_metrics = {metric_name: first_half.apply(lambda x: metric_func(x, step_risk_free_rate), axis=0) for metric_name, metric_func in metrics.items()}
        second_half_metrics = {metric_name: second_half.apply(lambda x: metric_func(x, step_risk_free_rate), axis=0) for metric_name, metric_func in metrics.items()}

        # Rank the trials in each half
        first_half_ranks = {metric_name: metrics_values.rank() for metric_name, metrics_values in first_half_metrics.items()}
        second_half_ranks = {metric_name: metrics_values.rank() for metric_name, metrics_values in second_half_metrics.items()}

        # Calculate the rank correlation for each metric
        for metric_name in metrics.keys():
            rank_corr, _ = kendalltau(first_half_ranks[metric_name], second_half_ranks[metric_name])
            rank_correlations[cv][metric_name] = rank_corr

    # Create the final DataFrame
    rank_corr_df = pd.DataFrame(rank_correlations)

    return rank_corr_df

def backtest_overfitting_simulation_model_complexity(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Any],
    step_risk_free_rate: float,
    n_jobs: int = 1
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Conducts a backtest overfitting simulation to compare the PBO and DSR values of each CV method for simple and complex models.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: Two DataFrames containing PBO and DSR values for each model and each CV method.
    """
    # Initialize DataFrames to store results
    pbo_df = pd.DataFrame(columns=['Combinatorial Purged'], index=models.keys())
    dsr_df = pd.DataFrame(columns=['Combinatorial Purged'], index=models.keys())

    # Create features
    volatility = daily_volatility_with_log_returns(prices, 100)
    filter_threshold = 1.8
    moelcules = cusum_filter_events_dynamic_threshold(np.log(prices), filter_threshold * volatility)
    vertical_barriers = vertical_barrier(prices, moelcules, 20)
    features = financial_features_backtest_overfitting_simulation(prices)

    for model_name, model in models.items():
        results = {
            'Combinatorial Purged' : [],
        }
        
        # Iterate over each strategy parameter combination
        strategy_parameters_keys, strategy_parameters_values = zip(*strategy_parameters.items())
        for strategy_parameters_value in itertools.product(*strategy_parameters_values):
            strategy_params = dict(zip(strategy_parameters_keys, strategy_parameters_value))
            try:
                strategy_sides = determine_strategy_side(prices, **strategy_params)

            except ValueError:    
                continue

            triple_barrier_events = meta_events(prices, moelcules, [0.5, 1.5], volatility, 0, 1, vertical_barriers, strategy_sides)
            labels = meta_labeling(triple_barrier_events, prices)
            sample_weights = sample_weight_absolute_return_meta_labeling(triple_barrier_events['End Time'], prices, moelcules)

            index = features.loc[moelcules].dropna().index.intersection(labels.dropna().index).intersection(sample_weights.dropna().index)
            data = features.loc[index]
            target = labels.loc[index]['Label']
            weights = sample_weights.loc[index]
            times = labels.loc[index]['End Time']

            cross_validators = {
                'Combinatorial Purged' : CrossValidatorController(
                    'combinatorialpurged',
                    n_splits=8,
                    n_test_groups=2,
                    times=times,
                    embargo=0.02
                ).cross_validator,
            }

            for cross_validator_type, cross_validator in cross_validators.items():
                predictions = cross_validator.backtest_predictions(model, data, target, weights, predict_probability=True, n_jobs=n_jobs)
                probabilities = pd.Series(np.vstack(list(map(lambda x: x[:, 1], predictions.values()))).mean(axis=0), times.index).dropna()
                positions = strategy_bet_sizing(prices.index, times.loc[probabilities.index], strategy_sides[probabilities.index], probabilities)
                strategy_log_returns = (np.log(prices).diff() * positions.shift()).dropna()
                results[cross_validator_type].append({
                    'Trial Info.' : {
                        'Strategy Parameters' : strategy_params,
                        'Model Name': model_name,
                        'Model Parameters': {},
                    },
                    'Returns': strategy_log_returns
                })

        cv_deflated_sr = {}
        cv_pbo = {}

        for cv, trials in results.items():
            performances = pd.concat([trial['Returns'] for trial in trials], axis=1)

            sharpe_ratios = performances.apply(lambda y: sharpe_ratio(y.values))    
            benchmark_sr = benchmark_sharpe_ratio(sharpe_ratios)
            best_strategy_index = sharpe_ratios.idxmax()
            cv_deflated_sr[cv] = probabilistic_sharpe_ratio(
                sharpe_ratios.loc[best_strategy_index], 
                benchmark_sr, len(performances), 
                ss.skew(performances[best_strategy_index]), 
                ss.kurtosis(performances[best_strategy_index]),
                return_test_statistic=True
            )

            pbo, logit_values = probability_of_backtest_overfitting(performances.values, risk_free_return=step_risk_free_rate, n_jobs=1)    
            cv_pbo[cv] = pbo

        for cv in results.keys():
            pbo_df.loc[model_name, cv] = cv_pbo[cv]
            dsr_df.loc[model_name, cv] = cv_deflated_sr[cv]

    return pbo_df, dsr_df

def noised_backtest_overfitting_simulation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    noise_scales: List[float],
    random_state: int = None,
    n_jobs: int = 1
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Conducts a noised backtest overfitting simulation to compare the new PBO/DSR values for different noise scales.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        noise_scales (List[float]): List of noise scale values to test.
        random_state (int): Seed for random number generator.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: Two DataFrames containing PBO and DSR values for each noise scale and each CV method.
    """
    # Initialize DataFrames to store results
    pbo_df = pd.DataFrame(columns=['Walk-Forward', 'K-Fold', 'Purged K-Fold', 'Combinatorial Purged'], index=noise_scales)
    dsr_df = pd.DataFrame(columns=['Walk-Forward', 'K-Fold', 'Purged K-Fold', 'Combinatorial Purged'], index=noise_scales)

    for noise_scale in noise_scales:
        # Perform the overall backtest overfitting simulation with the given noise scale
        cv_pbo, cv_deflated_sr = overall_backtest_overfitting_simulation(
            prices=prices,
            strategy_parameters=strategy_parameters,
            models=models,
            step_risk_free_rate=step_risk_free_rate,
            noise_scale=noise_scale,
            random_state=random_state,
            n_jobs=n_jobs
        )

        # Store the results in the DataFrames
        for cv in cv_pbo.keys():
            pbo_df.loc[noise_scale, cv] = cv_pbo[cv]
            dsr_df.loc[noise_scale, cv] = cv_deflated_sr[cv]

    return pbo_df, dsr_df

def overall_novel_methods_backtest_overfitting_simulation(
    prices: pd.Series, 
    strategy_parameters: Dict[str, Union[List[int], List[float], List[bool]]],
    models: Dict[str, Dict[str, Any]],
    step_risk_free_rate: float,
    noise_scale: float = 0.0,
    random_state: int = None,
    n_jobs: int = 1
) -> Tuple[Dict[str, float], Dict[str, float]]:
    """
    Conducts an overall backtest overfitting simulation to calculate the metrics for the novel CPCV methods.

    Args:
        prices (pd.Series): Time series of asset prices.
        strategy_parameters (dict): Parameters dictating trading strategy behavior.
        models (dict): A collection of machine learning models and their associated parameters.
        step_risk_free_rate (float): The risk-free rate used in the simulation for Sharpe ratio calculations.
        noise_scale (float): Scale of Gaussian noise to be added to the features.
        random_state (int): Seed for random number generator.
        n_jobs (int): The number of jobs to run in parallel for cross_validator.backtest_predictions.

    Returns:
        Tuple[Dict[str, float], Dict[str, float]]: A tuple containing two dictionaries, one for the Probability of Backtest Overfitting (PBO) and the other for the Deflated Sharpe Ratio (DSR), for each cross-validation method tested.
    """

    cross_validators = {
        'Combinatorial Purged' : CrossValidatorController(
            'combinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02
        ).cross_validator,
        'Bagged Combinatorial Purged' : CrossValidatorController(
            'baggedcombinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02,
            random_state=random_state
        ).cross_validator,
        'Adaptive Combinatorial Purged' : CrossValidatorController(
            'adaptivecombinatorialpurged',
            n_splits=8,
            n_test_groups=2,
            times=None,
            embargo=0.02,
        ).cross_validator,
    }

    results = backtest_overfitting_simulation_results(prices, strategy_parameters, models, cross_validators, noise_scale, random_state, n_jobs=n_jobs)


    cv_deflated_sr = {}
    cv_pbo = {}

    for cv, trials in results.items():
        performances = pd.concat([trial['Returns'] for trial in trials], axis=1)   

        sharpe_ratios = performances.apply(lambda y: sharpe_ratio(y.values))    
        benchmark_sr = benchmark_sharpe_ratio(sharpe_ratios)
        best_strategy_index = sharpe_ratios.idxmax()
        cv_deflated_sr[cv] = probabilistic_sharpe_ratio(
            sharpe_ratios.loc[best_strategy_index], 
            benchmark_sr, len(performances), 
            ss.skew(performances[best_strategy_index]), 
            ss.kurtosis(performances[best_strategy_index]),
            return_test_statistic=True
        )

        pbo, logit_values = probability_of_backtest_overfitting(performances.values, risk_free_return=step_risk_free_rate)    
        cv_pbo[cv] = pbo

    return cv_pbo, cv_deflated_sr    

def get_cpu_info():
    # Run the lscpu command
    result = subprocess.run(['lscpu'], stdout=subprocess.PIPE)
    # Decode the output from bytes to string
    lscpu_output = result.stdout.decode('utf-8')
    
    # Parse the lscpu output
    cpu_info = {}
    for line in lscpu_output.split('\n'):
        if line.strip():
            parts = line.split(':', 1)
            if len(parts) == 2:
                key, value = parts
                cpu_info[key.strip()] = value.strip()

    # Extract useful information
    useful_info = {
        "Architecture": cpu_info.get("Architecture"),
        "CPU op-mode(s)": cpu_info.get("CPU op-mode(s)"),
        "CPU(s)": cpu_info.get("CPU(s)"),
        "Thread(s) per core": cpu_info.get("Thread(s) per core"),
        "Core(s) per socket": cpu_info.get("Core(s) per socket"),
        "Socket(s)": cpu_info.get("Socket(s)"),
        "Vendor ID": cpu_info.get("Vendor ID"),
        "Model name": cpu_info.get("Model name"),
        "CPU MHz": cpu_info.get("CPU MHz"),
        "L1d cache": cpu_info.get("L1d cache"),
        "L1i cache": cpu_info.get("L1i cache"),
        "L2 cache": cpu_info.get("L2 cache"),
        "L3 cache": cpu_info.get("L3 cache"),
    }

    return useful_info

def format_cpu_info(cpu_info):
    report = (
        f"Architecture: {cpu_info['Architecture']}\n"
        f"CPU Operational Modes: {cpu_info['CPU op-mode(s)']}\n"
        f"Total CPUs: {cpu_info['CPU(s)']}\n"
        f"Threads per Core: {cpu_info['Thread(s) per core']}\n"
        f"Cores per Socket: {cpu_info['Core(s) per socket']}\n"
        f"Sockets: {cpu_info['Socket(s)']}\n"
        f"Vendor ID: {cpu_info['Vendor ID']}\n"
        f"Model Name: {cpu_info['Model name']}\n"
        f"CPU MHz: {cpu_info['CPU MHz']}\n"
        f"L1d Cache: {cpu_info['L1d cache']}\n"
        f"L1i Cache: {cpu_info['L1i cache']}\n"
        f"L2 Cache: {cpu_info['L2 cache']}\n"
        f"L3 Cache: {cpu_info['L3 cache']}\n"
    )
    return report

# Function to generate random data, target, weights, and times
def generate_random_data(n_samples: int, n_features: int) -> Tuple[pd.DataFrame, pd.Series, np.ndarray, pd.Series]:
    date_range = pd.date_range(start='1980-01-01', periods=n_samples, freq='1h')
    data = pd.DataFrame(np.random.randn(n_samples, n_features), columns=[f'feature_{i}' for i in range(n_features)], index=date_range)
    target = pd.Series(np.random.randint(0, 2, n_samples), index=date_range)
    weights = pd.Series(np.random.rand(n_samples), index=date_range)
    weights = weights / weights.sum()
    times = pd.Series(date_range + pd.DateOffset(hours=3), index=date_range)
    return data, target, weights, times

# Function to measure computational requirements
def measure_computational_requirements(cross_validator, model, data, target, weights, n_jobs: int = 1) -> Dict[str, Any]:
    start_time = time.time()
    mem_usage = memory_usage((cross_validator.backtest_predictions, (model, data, target, weights), {'predict_probability': True, 'n_jobs': n_jobs}), interval=0.1)
    end_time = time.time()
    return {
        'execution_time': end_time - start_time,
        'memory_usage': max(mem_usage)
    }

# Main function to measure computational requirements for all CV methods
def measure_all_cv_computational_requirements(
    cross_validators: Dict[str, Any],
    n_samples: int = 40 * 252, 
    n_features: int = 22, 
    n_jobs: int = 1,
    n_repeats: int = 30
) -> pd.DataFrame:
    # Generate random data, target, weights, and times
    data, target, weights, times = generate_random_data(n_samples, n_features)

    # Define the logistic regression model without regularization
    model = LogisticRegression(penalty=None, solver='lbfgs', max_iter=10000)
    
    # Measure computational requirements for each CV method
    results = {cv_name: {'execution_time': [], 'memory_usage': []} for cv_name in cross_validators.keys()}
    
    for _ in tqdm(range(n_repeats)):
        for cv_name, cross_validator in cross_validators.items():
            # print(f"Measuring computational requirements for {cv_name} (repeat {_ + 1}/{n_repeats})...")
            # Update cross-validator with times if required
            if 'times' in cross_validator.__dict__:
                cross_validator.times = times
            result = measure_computational_requirements(cross_validator, model, data, target, weights, n_jobs=n_jobs)
            results[cv_name]['execution_time'].append(result['execution_time'])
            results[cv_name]['memory_usage'].append(result['memory_usage'])

    # Calculate mean and standard deviation for each CV method
    results_summary = {}
    for cv_name, metrics in results.items():
        results_summary[cv_name] = {
            'execution_time_mean': np.mean(metrics['execution_time']),
            'execution_time_std': np.std(metrics['execution_time']),
            'memory_usage_mean': np.mean(metrics['memory_usage']),
            'memory_usage_std': np.std(metrics['memory_usage'])
        }

    # Convert results to DataFrame
    results_df = pd.DataFrame(results_summary).T

    return results_df

def measure_cpcv_parallelization(
    n_samples: int = 40 * 252, 
    n_features: int = 22, 
    n_repeats: int = 30,
    n_jobs_list: List[int] = range(1, 9)
) -> pd.DataFrame:
    # Generate random data, target, weights, and times
    data, target, weights, times = generate_random_data(n_samples, n_features)

    # Define the logistic regression model without regularization
    model = LogisticRegression(penalty=None, solver='lbfgs', max_iter=10000)

    # Define the CPCV cross-validator
    cpcv_cross_validator = CrossValidatorController('combinatorialpurged', n_splits=8, n_test_groups=2, times=times, embargo=0.02).cross_validator

    # Measure computational requirements with and without parallelization
    results = {}
    
    for n_jobs in tqdm(n_jobs_list):
        key = f'n_jobs_{n_jobs}'
        results[key] = {'execution_time': [], 'memory_usage': []}
        for _ in range(n_repeats):
            result = measure_computational_requirements(cpcv_cross_validator, model, data, target, weights, n_jobs=n_jobs)
            results[key]['execution_time'].append(result['execution_time'])
            results[key]['memory_usage'].append(result['memory_usage'])

    # Calculate mean and standard deviation for each setting
    results_summary = {}
    for key, metrics in results.items():
        results_summary[key] = {
            'execution_time_mean': np.mean(metrics['execution_time']),
            'execution_time_std': np.std(metrics['execution_time']),
            'memory_usage_mean': np.mean(metrics['memory_usage']),
            'memory_usage_std': np.std(metrics['memory_usage'])
        }

    # Convert results to DataFrame
    results_df = pd.DataFrame(results_summary).T

    return results_df

def measure_cpcv_scalability(
    sample_sizes: List[int],
    feature_sizes: List[int],
    n_repeats: int = 1,
    n_jobs: int = 1
) -> pd.DataFrame:
    # Define the logistic regression model without regularization
    model = LogisticRegression(penalty=None, solver='lbfgs', max_iter=10000)

    execution_times = pd.DataFrame(index=sample_sizes, columns=feature_sizes)
    memory_usages = pd.DataFrame(index=sample_sizes, columns=feature_sizes)

    for n_samples, n_features in itertools.product(sample_sizes, feature_sizes):
        execution_time_list = []
        memory_usage_list = []

        for _ in range(n_repeats):
            # Generate random data, target, weights, and times
            data, target, weights, times = generate_random_data(n_samples, n_features)

            # Define the CPCV cross-validator
            cpcv_cross_validator = CrossValidatorController('combinatorialpurged', n_splits=8, n_test_groups=2, times=times, embargo=0.02).cross_validator

            # Measure computational requirements
            result = measure_computational_requirements(cpcv_cross_validator, model, data, target, weights, n_jobs=n_jobs)
            execution_time_list.append(result['execution_time'])
            memory_usage_list.append(result['memory_usage'])

        # Calculate mean execution time and memory usage
        execution_times.loc[n_samples, n_features] = np.mean(execution_time_list)
        memory_usages.loc[n_samples, n_features] = np.mean(memory_usage_list)


    return execution_times, memory_usages



# THIS FILE NEEDS TO BE DELETED IN FUTURE VERSIONS


import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_samples
from scipy.linalg import block_diag
from sklearn.utils import check_random_state


def covariance_to_correlation(covariance: np.ndarray) -> np.ndarray:
    """
    Derive the correlation matrix from a covariance matrix.

    :param covariance: numpy ndarray
        The covariance matrix to convert to a correlation matrix.
    :return: numpy ndarray
        The correlation matrix derived from the covariance matrix.

    The conversion is done based on the following mathematical formula:
    correlation = covariance / (std_i * std_j)
    where std_i and std_j are the standard deviations of the i-th and j-th elements.
    """
    std = np.sqrt(np.diag(covariance))
    correlation = covariance / np.outer(std, std)
    correlation[correlation < -1], correlation[correlation > 1] = -1, 1
    return correlation


def cluster_kmeans_base(correlation: pd.DataFrame, number_clusters: int = 10, iterations: int = 10) -> tuple:
    """
    Apply the K-means clustering algorithm.

    :param correlation: pandas DataFrame
        The correlation matrix.
    :param number_clusters: int, optional
        The maximum number of clusters. Default is 10.
    :param iterations: int, optional
        The number of iterations to run the clustering. Default is 10.
    :return: tuple
        A tuple containing the sorted correlation matrix, cluster membership, and silhouette scores.

    This function is based on Snippet 4.1 from De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    """
    distance, silh = ((1 - correlation.fillna(0)) / 2.0) ** 0.5, pd.Series()
    for init in range(iterations):
        for i in range(2, number_clusters + 1):
            kmeans = KMeans(n_clusters=i, n_init=1)
            kmeans = kmeans.fit(distance)
            silh_ = silhouette_samples(distance, kmeans.labels_)
            statistic = (silh_.mean() / silh_.std(), silh.mean() / silh.std())
            if np.isnan(statistic[1]) or statistic[0] > statistic[1]:
                silh, kmeans = silh_, kmeans
    index_sorted = np.argsort(kmeans.labels_)
    correlation_sorted = correlation.iloc[index_sorted]
    correlation_sorted = correlation_sorted.iloc[:, index_sorted]
    clusters = {i: correlation.columns[np.where(kmeans.labels_ == i)[0]].tolist() for i in np.unique(kmeans.labels_)}
    silh = pd.Series(silh, index=distance.index)
    return correlation_sorted, clusters, silh


def make_new_outputs(correlation: pd.DataFrame, clusters: dict, clusters2: dict) -> tuple:
    """
    Merge two sets of clusters and derive new outputs.

    :param correlation: pandas DataFrame
        The correlation matrix.
    :param clusters: dict
        The first set of clusters.
    :param clusters2: dict
        The second set of clusters.
    :return: tuple
        A tuple containing the new correlation matrix, new cluster membership, and new silhouette scores.
    """
    clusters_new = {}
    for i in clusters.keys():
        clusters_new[len(clusters_new.keys())] = list(clusters[i])
    for i in clusters2.keys():
        clusters_new[len(clusters_new.keys())] = list(clusters2[i])
    index_new = [j for i in clusters_new for j in clusters_new[i]]
    correlation_new = correlation.loc[index_new, index_new]
    distance = ((1 - correlation.fillna(0)) / 2.0) ** 0.5
    labels_kmeans = np.zeros(len(distance.columns))
    for i in clusters_new.keys():
        index = [distance.index.get_loc(k) for k in clusters_new[i]]
        labels_kmeans[index] = i
    silh_new = pd.Series(silhouette_samples(distance, labels_kmeans), index=distance.index)
    return correlation_new, clusters_new, silh_new


def cluster_kmeans_top(correlation: pd.DataFrame, number_clusters: int = None, iterations: int = 10) -> tuple:
    """
    Apply the K-means clustering algorithm with hierarchical re-clustering.

    :param correlation: pandas DataFrame
        The correlation matrix.
    :param number_clusters: int, optional
        The maximum number of clusters. Default is None.
    :param iterations: int, optional
        The number of iterations to run the clustering. Default is 10.
    :return: tuple
        A tuple containing the sorted correlation matrix, cluster membership, and silhouette scores.

    This function is based on Snippet 4.2 from De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    """
    if number_clusters is None:
        number_clusters = correlation.shape[1] - 1
    correlation_sorted, clusters, silh = cluster_kmeans_base(
        correlation, min(number_clusters, correlation.shape[1] - 1), iterations
    )
    cluster_tstats = {i: np.mean(silh[clusters[i]]) / np.std(silh[clusters[i]]) for i in clusters.keys()}
    tstat_mean = sum(cluster_tstats.values()) / len(cluster_tstats)
    redo_clusters = [i for i in cluster_tstats.keys() if cluster_tstats[i] < tstat_mean]
    if len(redo_clusters) <= 1:
        return correlation_sorted, clusters, silh
    else:
        keys_redo = [j for i in redo_clusters for j in clusters[i]]
        correlation_temp = correlation.loc[keys_redo, keys_redo]
        tstat_mean = np.mean([cluster_tstats[i] for i in redo_clusters])
        correlation_sorted2, clusters2, silh2 = cluster_kmeans_top(
            correlation_temp, min(number_clusters, correlation_temp.shape[1] - 1), iterations
        )
        correlation_new, clusters_new, silh_new = make_new_outputs(
            correlation, {i: clusters[i] for i in clusters.keys() if i not in redo_clusters}, clusters2
        )
        new_tstat_mean = np.mean(
            [np.mean(silh_new[clusters_new[i]]) / np.std(silh_new[clusters_new[i]]) for i in clusters_new.keys()]
        )
        if new_tstat_mean <= tstat_mean:
            return correlation_sorted, clusters, silh
        else:
            return correlation_new, clusters_new, silh_new


import numpy as np
from scipy.linalg import block_diag
from sklearn.utils import check_random_state


def random_covariance_sub(number_observations: int, number_columns: int, sigma: float, random_state=None) -> np.ndarray:
    """
    Compute a sub covariance matrix.

    Generates a covariance matrix based on random data.

    :param number_observations: Number of observations.
    :param number_columns: Number of columns.
    :param sigma: Sigma for normal distribution.
    :param random_state: Random state for reproducibility.
    :return: Sub covariance matrix.

    .. note:: Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
       Methodology: Snipet 4.3, Page 61.
    """
    domain = check_random_state(random_state)
    if number_columns == 1:
        return np.ones((1, 1))
    data = domain.normal(size=(number_observations, 1))  # generate data
    data = np.repeat(data, number_columns, axis=1)  # repeat data
    data += domain.normal(scale=sigma, size=data.shape)  # add data
    covariance = np.cov(data, rowvar=False)  # covariance of data
    return covariance


def random_block_covariance(
    number_columns: int, number_blocks: int, block_size_min: int = 1, sigma: float = 1.0, random_state=None
) -> np.ndarray:
    """
    Compute a random block covariance matrix.

    Generates a block random covariance matrix by combining multiple sub covariance matrices.

    :param number_columns: Number of columns.
    :param number_blocks: Number of blocks.
    :param block_size_min: Minimum size of block.
    :param sigma: Sigma for normal distribution.
    :param random_state: Random state for reproducibility.
    :return: Block random covariance matrix.

    .. note:: Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
       Methodology: Snipet 4.3, Page 61.
    """
    domain = check_random_state(random_state)
    parts = domain.choice(
        range(1, number_columns - (block_size_min - 1) * number_blocks), number_blocks - 1, replace=False
    )
    parts.sort()
    parts = np.append(parts, number_columns - (block_size_min - 1) * number_blocks)
    parts = np.append(parts[0], np.diff(parts)) - 1 + block_size_min
    cov = None
    for column in parts:
        this_covariance = random_covariance_sub(
            int(max(column * (column + 1) / 2.0, 100)), column, sigma, random_state=domain
        )  # sub covariance
        if cov is None:
            cov = this_covariance.copy()  # copy covariance matrix
        else:
            cov = block_diag(cov, this_covariance)  # block diagram covariance matrix
    return cov


import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_samples
from sklearn.utils import check_random_state


def random_block_correlation(
    number_columns: int, number_blocks: int, random_state=None, block_size_min: int = 1
) -> pd.DataFrame:
    """
    Compute a random block correlation matrix.

    Generates a block random correlation matrix by adding two block random covariance matrices
    and converting them to a correlation matrix.

    :param number_columns: Number of columns.
    :param number_blocks: Number of blocks.
    :param random_state: Random state for reproducibility.
    :param block_size_min: Minimum size of block.
    :return: Block random correlation matrix.

    .. note:: Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
       Methodology: Snipet 4.3, Page 61.
    """
    domain = check_random_state(random_state)
    covariance1 = random_block_covariance(
        number_columns, number_blocks, block_size_min=block_size_min, sigma=0.5, random_state=domain
    )
    covariance2 = random_block_covariance(
        number_columns, 1, block_size_min=block_size_min, sigma=1.0, random_state=domain
    )
    covariance1 += covariance2
    correlation = cov_to_corr(covariance1)
    correlation = pd.DataFrame(correlation)
    return correlation


def cov_to_corr(covariance: np.ndarray) -> np.ndarray:
    r"""
    Convert a covariance matrix to a correlation matrix.

    :param covariance: Covariance matrix.
    :return: Correlation matrix.

    .. math::
        correlation_{ij} = \\frac{covariance_{ij}}{\\sqrt{covariance_{ii} \cdot covariance_{jj}}}
    """
    std = np.sqrt(np.diag(covariance))
    correlation = covariance / np.outer(std, std)
    correlation[correlation < -1] = -1
    correlation[correlation > 1] = 1
    return correlation


def cluster_kmeans_base(
    correlation: pd.DataFrame, number_clusters: int = 10, iterations: int = 10
) -> (pd.DataFrame, dict, pd.Series):
    """
    Perform KMeans clustering on a correlation matrix.

    :param correlation: Correlation matrix.
    :param number_clusters: Number of clusters, default is 10.
    :param iterations: Number of iterations, default is 10.
    :return: Sorted correlation matrix, clusters, silhouette scores.

    .. note::
        The silhouette score is a measure of how similar an object is to its own cluster compared to other clusters.
    """
    distance = ((1 - correlation.fillna(0)) / 2.0) ** 0.5
    silh = pd.Series()

    for init in range(iterations):
        for i in range(2, number_clusters + 1):
            kmeans_ = KMeans(n_clusters=i, n_init=1)
            kmeans_ = kmeans_.fit(distance)
            silh_ = silhouette_samples(distance, kmeans_.labels_)
            statistic = (silh_.mean() / silh_.std(), silh.mean() / silh.std())

            if np.isnan(statistic[1]) or statistic[0] > statistic[1]:
                silh, kmeans = silh_, kmeans_

    index_sorted = np.argsort(kmeans.labels_)
    correlation_sorted = correlation.iloc[index_sorted]
    correlation_sorted = correlation_sorted.iloc[:, index_sorted]

    clusters = {i: correlation.columns[np.where(kmeans.labels_ == i)[0]].tolist() for i in np.unique(kmeans.labels_)}

    silh = pd.Series(silh, index=distance.index)
    return correlation_sorted, clusters, silh


# Example usage:
if __name__ == "__main__":
    correlation_matrix = random_block_correlation(number_columns=20, number_blocks=5)
    sorted_corr, cluster_dict, silh_scores = cluster_kmeans_base(correlation_matrix)
    print(sorted_corr)
    print(cluster_dict)
    print(silh_scores)



"""
Functions for calculating metrics related to test-set overfitting,
including the expected maximum Sharpe Ratio and Type I/II errors.
"""

from typing import List
import numpy as np
import pandas as pd
import scipy.stats as ss
from scipy.stats import norm

def expected_max_sharpe_ratio(
    n_trials: int, mean_sharpe_ratio: float, std_sharpe_ratio: float
) -> float:
    r"""
    Calculate the expected maximum Sharpe Ratio from N trials.

    Uses the formula for the expected maximum of N standard normal variables:
    .. math::
        E[\max(SR)] = \mu + \sigma \left[ (1 - \gamma) \Phi^{-1}(1 - \frac{1}{N}) +
                       \gamma \Phi^{-1}(1 - (N e)^{-1}) \right]

    Where:
    - \(\mu\) is the mean_sharpe_ratio
    - \(\sigma\) is the std_sharpe_ratio
    - \(\gamma\) is the Euler-Mascheroni constant
    - \(\Phi^{-1}\) is the inverse CDF of the standard normal distribution
    - \(N\) is n_trials

    Parameters
    ----------
    n_trials : int
        Number of trials (e.g., number of strategies tested).
    mean_sharpe_ratio : float
        The mean Sharpe Ratio across all trials.
    std_sharpe_ratio : float
        The standard deviation of Sharpe Ratios across all trials.

    Returns
    -------
    float
        The expected maximum Sharpe Ratio.
    """
    if n_trials == 0:
        return 0.0
    if n_trials == 1:
        return mean_sharpe_ratio
        
    euler_gamma = 0.5772156649

    term1 = (1 - euler_gamma) * norm.ppf(1.0 - 1.0 / n_trials)
    term2 = euler_gamma * norm.ppf(1.0 - (n_trials * np.e) ** -1)

    expected_max_sr = mean_sharpe_ratio + std_sharpe_ratio * (term1 + term2)

    return expected_max_sr

def generate_max_sharpe_ratios(
    n_sims: int,
    n_trials_list: List[int],
    std_sharpe_ratio: float,
    mean_sharpe_ratio: float,
) -> pd.DataFrame:
    """
    Generate a DataFrame of maximum Sharpe Ratios from simulations.

    Parameters
    ----------
    n_sims : int
        Number of simulations to run for each `n_trials`.
    n_trials_list : List[int]
        A list of N-trials (e.g., [10, 50, 100] strategies).
    std_sharpe_ratio : float
        The standard deviation of Sharpe Ratios.
    mean_sharpe_ratio : float
        The mean of Sharpe Ratios.

    Returns
    -------
    pd.DataFrame
        A long-format DataFrame with columns ['max_SR', 'n_trials'].
    """
    rng = np.random.default_rng()
    output_list = []

    for n_trials in n_trials_list:
        # Generate all simulations for this n_trials
        sr_sims = rng.normal(
            loc=0.0, scale=1.0, size=(n_sims, n_trials)
        )
        
        # Normalize (z-score) each simulation row
        sr_sims = (sr_sims - sr_sims.mean(axis=1, keepdims=True)) / \
                  sr_sims.std(axis=1, keepdims=True)
        
        # Scale by target mean and std
        sr_sims = mean_sharpe_ratio + sr_sims * std_sharpe_ratio

        # Get max SR for each simulation
        max_sr = sr_sims.max(axis=1)
        
        output_temp = pd.DataFrame({'max_SR': max_sr, 'n_trials': n_trials})
        output_list.append(output_temp)

    return pd.concat(output_list, ignore_index=True)


def mean_std_error(
    n_sims0: int,
    n_sims1: int,
    n_trials: List[int],
    std_sharpe_ratio: float = 1.0,
    mean_sharpe_ratio: float = 0.0,
) -> pd.DataFrame:
    """
    Calculate the mean and standard deviation of the estimation error.

    This function compares the analytical E[max{SR}] with the simulated
    average max{SR} over `n_sims1` repetitions.

    Parameters
    ----------
    n_sims0 : int
        Number of max{SR} simulations used to estimate E[max{SR}] (inner loop).
    n_sims1 : int
        Number of errors on which to compute the std dev (outer loop).
    n_trials : List[int]
        List of numbers of trials (e.g., [10, 20, 50]).
    std_sharpe_ratio : float, default=1.0
        Standard deviation of Sharpe Ratios.
    mean_sharpe_ratio : float, default=0.0
        Mean of Sharpe Ratios.

    Returns
    -------
    pd.DataFrame
        DataFrame indexed by 'nTrials' with columns ['meanErr', 'stdErr'].
    """
    # 1. Analytical E[max{SR}]
    expected_sr = pd.Series(
        {
            n: expected_max_sharpe_ratio(n, mean_sharpe_ratio, std_sharpe_ratio)
            for n in n_trials
        },
        name="E[max{SR}]",
    )
    expected_sr.index.name = "nTrials"

    error_list = []
    
    # 2. Run n_sims1 experiments
    for _ in range(int(n_sims1)):
        # 3. Generate simulated max SRs
        simulated_sr = generate_max_sharpe_ratios(
            n_sims=n_sims0,
            n_trials=n_trials,
            mean_sharpe_ratio=mean_sharpe_ratio,
            std_sharpe_ratio=std_sharpe_ratio,
        )
        # 4. Average simulated max SRs
        avg_simulated_sr = simulated_sr.groupby("nTrials")["max_SR"].mean()
        
        # 5. Calculate error
        error = avg_simulated_sr / expected_sr - 1.0
        error_list.append(error.rename('error'))

    all_errors = pd.concat(error_list, axis=1).T

    # 6. Compute mean and std of errors
    output = pd.DataFrame({
        "meanErr": all_errors.mean(),
        "stdErr": all_errors.std(),
    })

    return output


def estimated_sharpe_ratio_z_statistics(
    sharpe_ratio: float,
    t: int,
    true_sharpe_ratio: float = 0.0,
    skew: float = 0.0,
    kurt: int = 3,
) -> float:
    r"""
    Calculate the Z-statistic for an estimated Sharpe Ratio.

    This is the test statistic used in the Probabilistic Sharpe Ratio.

    .. math::
        Z = \frac{(\hat{SR} - SR_0) \sqrt{T - 1}}
                 {\sqrt{1 - S \hat{SR} + \frac{K - 1}{4} \hat{SR}^2}}

    Parameters
    ----------
    sharpe_ratio : float
        Estimated Sharpe Ratio (\(\hat{SR}\)).
    t : int
        Number of observations (T).
    true_sharpe_ratio : float, default=0.0
        The null hypothesis Sharpe Ratio (\(SR_0\)).
    skew : float, default=0.0
        Skewness of returns (S).
    kurt : int, default=3
        Kurtosis of returns (K).

    Returns
    -------
    float
        The calculated Z-statistic.
    """
    denominator = (
        1 - skew * sharpe_ratio + (kurt - 1) / 4.0 * sharpe_ratio**2
    )
    if denominator <= 0:
        return np.nan
        
    z = (sharpe_ratio - true_sharpe_ratio) * np.sqrt(t - 1)
    z /= np.sqrt(denominator)

    return z


def strategy_type1_error_probability(z: float, k: int = 1) -> float:
    r"""
    Calculate the Type I error probability (alpha) for multiple tests.

    This is the probability of at least one false positive (rejecting
    a true null) when conducting `k` independent tests.

    .. math::
        \alpha_k = 1 - (1 - \alpha)^k

    Where \(\alpha = N(-z)\) is the Type I error for a single test.

    Parameters
    ----------
    z : float
        Z-statistic for the significance threshold (e.g., 1.96).
    k : int, default=1
        Number of independent tests.

    Returns
    -------
    float
        The family-wise Type I error rate.
    """
    alpha_single_test = ss.norm.cdf(-z)
    alpha_k = 1 - (1 - alpha_single_test) ** k
    return alpha_k


def theta_for_type2_error(
    sharpe_ratio: float,
    t: int,
    true_sharpe_ratio: float,
    skew: float = 0.0,
    kurt: int = 3,
) -> float:
    r"""
    Calculate the \(\theta\) parameter for Type II error probability.

    .. math::
        \theta = \frac{SR_{True} \sqrt{T - 1}}
                      {\sqrt{1 - S \hat{SR} + \frac{K - 1}{4} \hat{SR}^2}}

    Parameters
    ----------
    sharpe_ratio : float
        The estimated Sharpe Ratio (\(\hat{SR}\)).
    t : int
        Number of observations (T).
    true_sharpe_ratio : float
        The true Sharpe Ratio (\(SR_{True}\)) (the alternative hypothesis).
    skew : float, default=0.0
        Skewness of returns (S).
    kurt : int, default=3
        Kurtosis of returns (K).

    Returns
    -------
    float
        The \(\theta\) parameter.
    """
    denominator = (
        1 - skew * sharpe_ratio + (kurt - 1) / 4.0 * sharpe_ratio**2
    )
    if denominator <= 0:
        return np.nan
        
    theta = true_sharpe_ratio * np.sqrt(t - 1)
    theta /= np.sqrt(denominator)
    return theta


def strategy_type2_error_probability(
    alpha_k: float, k: int, theta: float
) -> float:
    r"""
    Calculate the Type II error probability (beta) for multiple tests.

    This is the probability of failing to reject a false null hypothesis.

    .. math::
        Z_{\alpha} = \Phi^{-1}((1 - \alpha_k)^{1/k})
        \beta = N(Z_{\alpha} - \theta)

    Parameters
    ----------
    alpha_k : float
        The family-wise Type I error rate.
    k : int
        Number of independent tests.
    theta : float
        The \(\theta\) parameter from `theta_for_type2_error`.

    Returns
    -------
    float
        The Type II error probability (\(\beta\)).
    """
    z_alpha = ss.norm.ppf((1 - alpha_k) ** (1.0 / k))
    beta = ss.norm.cdf(z_alpha - theta)
    return beta


"""
Calculates the theoretical accuracy of an ensemble of classifiers.
"""

from math import comb, floor
from typing import Union
import numpy as np

def bagging_classifier_accuracy(
    N: int, p: float, k: int = 2
) -> Union[float, np.float64]:
    r"""
    Calculate the accuracy of a bagging classifier.

    This function calculates the probability that the majority vote of
    an ensemble of `N` classifiers, each with an accuracy `p`,
    will be correct.

    The formula computes the sum of the binomial probabilities for a
    correct majority vote:
    .. math::
        P(\text{correct}) = \sum_{i=\lfloor N/k \rfloor + 1}^{N}
                           \binom{N}{i} p^i (1-p)^{N-i}

    Reference:
        De Prado, M. (2018) Advances in financial machine learning.
        Methodology: page 96, "Improved Accuracy" section.

    Parameters
    ----------
    N : int
        Number of independent classifiers. Must be odd (or will be incremented).
    p : float
        Probability of a single classifier being correct (accuracy).
    k : int, default=2
        Number of classes. Assumes `k=2` for binary classification.

    Returns
    -------
    float
        The bagging classifier's overall accuracy.
    """
    if N % 2 == 0:
        # N must be odd to avoid ties for k=2
        N += 1
        
    # Majority threshold
    majority_threshold = floor(N / k) + 1
    
    # Sum the probabilities from the first majority (threshold) to N
    prob_sum = sum(
        comb(N, i) * (p**i) * ((1 - p) ** (N - i))
        for i in range(majority_threshold, N + 1)
    )

    return prob_sum


"""
Functions for calculating strategy risk metrics, such as implied
precision, binomial Sharpe ratio, and probability of failure.
"""

from typing import Tuple
import numpy as np
import scipy.stats as ss
import sympy
from sympy import symbols, factor

def sharpe_ratio_trials(p: float, n_run: int) -> Tuple[float, float, float]:
    r"""
    Simulate binomial trials to estimate mean, std dev, and Sharpe ratio.

    Each trial is a binomial outcome (1 or -1) with probability `p`.

    .. math::
        \text{outcome} = \begin{cases} 1 & \text{with probability } p \\
                                     -1 & \text{with probability } 1-p
                       \end{cases}
        SR = \frac{\mu}{\sigma}

    Parameters
    ----------
    p : float
        Probability of success (outcome 1).
    n_run : int
        Number of simulation runs (trials).

    Returns
    -------
    Tuple[float, float, float]
        (mean_outcome, std_outcome, sharpe_ratio)
    """
    outcomes = np.random.binomial(n=1, p=p, size=n_run) * 2 - 1
    mean_outcome = np.mean(outcomes)
    std_outcome = np.std(outcomes)
    sharpe_ratio = mean_outcome / std_outcome if std_outcome > 0 else 0.0

    return mean_outcome, std_outcome, sharpe_ratio


def target_sharpe_ratio_symbolic() -> sympy.Expr:
    r"""
    Calculate the variance of a binomial trial symbolically.

    The variance is:
    .. math::
        Var = E[X^2] - (E[X])^2
        Var = (p u^2 + (1 - p) d^2) - (p u + (1 - p) d)^2

    This function simplifies this expression using SymPy.

    Returns
    -------
    sympy.Expr
        The symbolic expression for variance: `p*(1 - p)*(d - u)**2`
    """
    p, u, d = symbols("p u d")

    m2 = p * u**2 + (1 - p) * d**2
    m1 = p * u + (1 - p) * d
    v = m2 - m1**2

    return factor(v)


def implied_precision(
    stop_loss: float,
    profit_taking: float,
    frequency: float,
    target_sharpe_ratio: float,
) -> float:
    r"""
    Calculate the implied precision (probability) to achieve a target SR.

    Solves the quadratic equation for `p` from the binomial SR formula:
    .. math::
        a = (f + S^2) (pt - sl)^2
        b = (2 f sl - S^2 (pt - sl)) (pt - sl)
        c = f sl^2
        p = \frac{-b + \sqrt{b^2 - 4ac}}{2a}

    Where:
    - \(f\) is the frequency (bets per year)
    - \(S\) is the target_sharpe_ratio
    - \(pt\) is the profit_taking
    - \(sl\) is the stop_loss (as a positive value)

    Parameters
    ----------
    stop_loss : float
        Stop-loss threshold (positive value, e.g., 0.02 for 2%).
    profit_taking : float
        Profit-taking threshold (positive value, e.g., 0.04 for 4%).
    frequency : float
        Number of bets per year.
    target_sharpe_ratio : float
        Target annual Sharpe ratio.

    Returns
    -------
    float
        The required precision (probability of success).
    """
    a = (frequency + target_sharpe_ratio**2) * (profit_taking - stop_loss) ** 2
    b = (
        (2 * frequency * stop_loss - target_sharpe_ratio**2 * (profit_taking - stop_loss))
        * (profit_taking - stop_loss)
    )
    c = frequency * stop_loss**2
    
    discriminant = b**2 - 4 * a * c
    if discriminant < 0:
        return np.nan # No real solution

    precision = (-b + np.sqrt(discriminant)) / (2.0 * a)
    return precision


def bin_frequency(
    stop_loss: float,
    profit_taking: float,
    precision: float,
    target_sharpe_ratio: float,
) -> float:
    r"""
    Calculate the frequency needed to achieve a target SR.

    .. math::
        f = \frac{S^2 (pt - sl)^2 p (1 - p)}{((pt - sl) p - sl)^2}
    
    Note: The original formula had `((pt - sl) * p + sl)` in the denominator,
    which corresponds to `sl` being negative (a loss). This
    implementation assumes `stop_loss` is a positive value, so the
    expected return is `(pt - sl) * p - sl * (1-p) = p*pt - p*sl - sl + p*sl = p*pt - sl`.
    Wait, the original `binomial_sharpe_ratio` uses `(pt - sl) * p + sl`.
    This implies `sl` is treated as a negative number.
    Let's assume `stop_loss` is a positive threshold, so `d = -stop_loss`.
    E[R] = p*pt + (1-p)*(-sl) = p*(pt+sl) - sl
    Stdev = (pt - (-sl)) * sqrt(p(1-p)) = (pt+sl)*sqrt(p(1-p))
    SR = (p*(pt+sl) - sl) / ((pt+sl)*sqrt(p(1-p)))
    Let's stick to the formula provided in `binomial_sharpe_ratio`.

    .. math::
        f = \frac{S^2 (pt - sl)^2 p (1-p)}{((pt - sl) p + sl)^2}

    Parameters
    ----------
    stop_loss : float
        Stop-loss threshold (e.g., -0.02 for 2% loss).
    profit_taking : float
        Profit-taking threshold (e.g., 0.04 for 4% gain).
    precision : float
        The precision (probability of a profitable trade).
    target_sharpe_ratio : float
        The target annual Sharpe ratio.

    Returns
    -------
    float
        The required number of bets per year (frequency).
    """
    if precision <= 0 or precision >= 1:
        return np.inf

    numerator = (
        (target_sharpe_ratio * (profit_taking - stop_loss)) ** 2
        * precision
        * (1 - precision)
    )
    denominator = ((profit_taking - stop_loss) * precision + stop_loss) ** 2
    
    if denominator == 0:
        return np.inf

    return numerator / denominator


def binomial_sharpe_ratio(
    stop_loss: float,
    profit_taking: float,
    frequency: float,
    probability: float,
) -> float:
    r"""
    Calculate the Sharpe Ratio for a binary outcome strategy.

    .. math::
        E[R] = p \cdot pt + (1-p) \cdot sl
        \sigma[R] = (pt - sl) \sqrt{p(1-p)}
        SR_{trade} = \frac{E[R]}{\sigma[R]}
        SR_{annual} = SR_{trade} \times \sqrt{f}

    Parameters
    ----------
    stop_loss : float
        Stop-loss outcome (negative value, e.g., -0.02).
    profit_taking : float
        Profit-taking outcome (positive value, e.g., 0.04).
    frequency : float
        Number of bets per year.
    probability : float
        Probability of success (profit_taking).

    Returns
    -------
    float
        The annualized Sharpe Ratio.
    """
    expected_return = (profit_taking * probability) + (stop_loss * (1 - probability))
    
    p = probability
    stdev_return = (profit_taking - stop_loss) * np.sqrt(p * (1 - p))

    if stdev_return == 0:
        return 0.0 if expected_return == 0 else np.inf * np.sign(expected_return)

    sr_trade = expected_return / stdev_return
    sr_annual = sr_trade * np.sqrt(frequency)
    
    return sr_annual


def mix_gaussians(
    mu1: float,
    mu2: float,
    sigma1: float,
    sigma2: float,
    probability: float,
    n_obs: int,
) -> np.ndarray:
    """
    Generate a mixture of two Gaussian-distributed outcomes.

    Parameters
    ----------
    mu1 : float
        Mean of the first Gaussian (e.g., winning trades).
    mu2 : float
        Mean of the second Gaussian (e.g., losing trades).
    sigma1 : float
        Standard deviation of the first Gaussian.
    sigma2 : float
        Standard deviation of the second Gaussian.
    probability : float
        Probability of drawing from the first Gaussian (e.g., prob of win).
    n_obs : int
        Total number of observations (trades) to generate.

    Returns
    -------
    np.ndarray
        Array of generated outcomes, shuffled.
    """
    n_obs1 = int(n_obs * probability)
    n_obs2 = n_obs - n_obs1
    
    returns1 = np.random.normal(mu1, sigma1, size=n_obs1)
    returns2 = np.random.normal(mu2, sigma2, size=n_obs2)

    returns = np.append(returns1, returns2, axis=0)
    np.random.shuffle(returns)

    return returns


def failure_probability(
    returns: np.ndarray, frequency: float, target_sharpe_ratio: float
) -> float:
    """
    Calculate the probability that the strategy may fail.

    This compares the observed precision (from returns) with the
    implied precision required to meet the target SR.

    Parameters
    ----------
    returns : np.ndarray
        Array of returns from a strategy.
    frequency : float
        Number of bets per year (observed).
    target_sharpe_ratio : float
        The target annual Sharpe ratio.

    Returns
    -------
    float
        Calculated failure probability (a Z-score CDF).
    """
    positive_returns = returns[returns > 0]
    negative_returns = returns[returns <= 0]

    if len(positive_returns) == 0 or len(negative_returns) == 0:
        return 0.0 # Cannot calculate

    profit_taking = positive_returns.mean()
    stop_loss = negative_returns.mean() # This will be negative
    
    # Observed precision
    observed_precision = positive_returns.shape[0] / float(returns.shape[0])

    # Required precision
    required_precision = implied_precision(
        abs(stop_loss), profit_taking, frequency, target_sharpe_ratio
    )
    
    if np.isnan(required_precision):
        return 1.0 # Cannot achieve target SR

    # Probability that observed_precision < required_precision
    # This is a test on a proportion
    p_var = observed_precision * (1 - observed_precision)
    if p_var == 0:
        return 0.0 if observed_precision >= required_precision else 1.0
        
    p_std = np.sqrt(p_var / returns.shape[0]) # Std dev of the proportion
    
    z_score = (observed_precision - required_precision) / p_std
    risk = ss.norm.cdf(z_score) # Prob of being <= required_precision

    return risk

def calculate_strategy_risk(
    mu1: float,
    mu2: float,
    sigma1: float,
    sigma2: float,
    probability: float,
    n_obs: int,
    frequency: float,
    target_sharpe_ratio: float,
) -> float:
    """
    Run a simulation to calculate the strategy risk.

    Parameters
    ----------
    mu1 : float
        Mean of winning trades.
    mu2 : float
        Mean of losing trades.
    sigma1 : float
        Standard deviation of winning trades.
    sigma2 : float
        Standard deviation of losing trades.
    probability : float
        Probability of a winning trade.
    n_obs : int
        Number of observations (trades) in the simulation.
    frequency : float
        Number of bets per year.
    target_sharpe_ratio : float
        Target annual Sharpe ratio.

    Returns
    -------
    float
        Calculated probability of strategy failure.
    """
    returns = mix_gaussians(
        mu1, mu2, sigma1, sigma2, probability, n_obs
    )
    probability_fail = failure_probability(
        returns, frequency, target_sharpe_ratio
    )
    
    print(f"Probability that strategy will fail: {probability_fail:.2%}")
    return probability_fail


"""
RiskLabAI: Financial AI with Python

This library is a Python-based implementation of advanced
methods for quantitative finance and financial machine learning,
based on the work of Marcos López de Prado.

Sub-modules
-----------
backtest
    Tools for robust backtesting, including PBO, PSR, and
    advanced cross-validation (PurgedKFold, CPCV).
cluster
    Algorithms for portfolio clustering (ONC).
controller
    High-level controllers for bar generation and data processing.
data
    Core modules for data processing, including bar generation,
    labeling, denoising, and synthetic data.
ensemble
    Functions related to ensemble methods.
features
    Modules for feature generation and importance analysis
    (MDI, MDA, SFI, structural breaks, entropy, microstructure).
hpc
    Utilities for high-performance computing and parallel processing.
optimization
    Portfolio optimization algorithms, including HRP and NCO.
pde
    Implementation of a Deep BSDE solver for PDEs.
utils
    Common helper functions, constants, and plotting utilities.
"""

# Import all sub-packages to make them available
from . import backtest
from . import cluster
from . import controller
from . import data
from . import ensemble
from . import features
from . import hpc
from . import optimization
from . import pde
from . import utils

__version__ = "0.0.93"

__all__ = [
    "backtest",
    "cluster",
    "controller",
    "data",
    "ensemble",
    "features",
    "hpc",
    "optimization",
    "pde",
    "utils",
    "__version__",
]


"""
A helper function to apply a consistent, dark-themed layout
to Plotly figures.
"""

import plotly.graph_objects as go
from typing import Optional

def update_figure_layout(
    fig: go.Figure,
    title: str,
    xaxis_title: str,
    yaxis_title: str,
    legend_x: float = 1.0,
    legend_y: float = 1.0,
) -> go.Figure:
    """
    Apply a standardized dark-theme layout to a Plotly figure.

    This function modifies the figure in-place.

    Parameters
    ----------
    fig : go.Figure
        The Plotly figure object to update.
    title : str
        The title for the chart.
    xaxis_title : str
        The title for the x-axis.
    yaxis_title : str
        The title for the y-axis.
    legend_x : float, default=1.0
        The x-position of the legend.
    legend_y : float, default=1.0
        The y-position of the legend.

    Returns
    -------
    go.Figure
        The modified figure.
    """
    fig.update_layout(
        title=title,
        xaxis_title=xaxis_title,
        yaxis_title=yaxis_title,
        template="plotly_dark",
        plot_bgcolor="rgba(0,0,0,0)",  # Transparent background
        paper_bgcolor="rgba(0,0,0,0)", # Transparent background
        legend=dict(
            x=legend_x,
            y=legend_y,
            xanchor="auto",
            yanchor="auto",
        ),
    )
    return fig


"""
RiskLabAI Feature Importance Module

Provides a framework for calculating feature importance using
various methods (MDI, MDA, SFI) and clustering.
"""

from .feature_importance_strategy import FeatureImportanceStrategy
from .feature_importance_factory import FeatureImportanceFactory
from .feature_importance_controller import FeatureImportanceController

from .feature_importance_mdi import FeatureImportanceMDI
from .clustered_feature_importance_mdi import ClusteredFeatureImportanceMDI
from .feature_importance_mda import FeatureImportanceMDA
from .clustered_feature_importance_mda import ClusteredFeatureImportanceMDA
from .feature_importance_sfi import FeatureImportanceSFI

# Other utilities from the original __init__
# (Assuming they exist, as they were not provided)
# from .clustering import *
# from .generate_synthetic_data import *
# from .orthogonal_features import *
# from .weighted_tau import *
# from . import swefi

__all__ = [
    # Core Pattern
    "FeatureImportanceStrategy",
    "FeatureImportanceFactory",
    "FeatureImportanceController",
    
    # Strategy Implementations
    "FeatureImportanceMDI",
    "ClusteredFeatureImportanceMDI",
    "FeatureImportanceMDA",
    "ClusteredFeatureImportanceMDA",
    "FeatureImportanceSFI",
]


"""
Concrete implementations of Run Bars:
- ExpectedRunBars
- FixedRunBars
"""

from typing import Union, Iterable, List, Optional, Tuple
import numpy as np

from RiskLabAI.data.structures.abstract_run_bars import AbstractRunBars
from RiskLabAI.utils.constants import *

# Assuming ewma is in utils
try:
    from RiskLabAI.utils.ewma import ewma 
except ImportError:
    # Fallback if ewma is not in utils (as seen in older files)
    def ewma(array: np.ndarray, window: int) -> np.ndarray:
        """Placeholder EWMA function."""
        if array.size == 0:
            return np.array([np.nan])
        return pd.Series(array).ewm(span=window).mean().values


class ExpectedRunBars(AbstractRunBars):
    """
    Concrete class for Run Bars with a dynamic, EWMA-based
    Expected Number of Ticks (E[T]).
    """

    def __init__(
        self,
        bar_type: str,
        window_size_for_expected_n_ticks_estimation: int,
        initial_estimate_of_expected_n_ticks_in_bar: int,
        window_size_for_expected_imbalance_estimation: int,
        expected_ticks_number_bounds: Optional[Tuple[float, float]] = None,
        analyse_thresholds: bool = False,
    ):
        """
        Constructor.

        Parameters
        ----------
        bar_type : str
            e.g., 'dollar_run', 'tick_run'.
        window_size_for_expected_n_ticks_estimation : int
            Window size for EWMA of E[T].
        initial_estimate_of_expected_n_ticks_in_bar : int
            Initial guess for E[T].
        window_size_for_expected_imbalance_estimation : int
            Window size for EWMA of E[theta].
        expected_ticks_number_bounds : Tuple[float, float], optional
            (Lower, Upper) bounds to clamp E[T].
        analyse_thresholds : bool, default=False
            If True, store threshold data for analysis.
        """
        super().__init__(
            bar_type,
            window_size_for_expected_n_ticks_estimation,
            window_size_for_expected_imbalance_estimation,
            initial_estimate_of_expected_n_ticks_in_bar,
            analyse_thresholds,
        )

        if expected_ticks_number_bounds is None:
            self.expected_ticks_number_lower_bound = 0.0
            self.expected_ticks_number_upper_bound = np.inf
        else:
            self.expected_ticks_number_lower_bound = (
                expected_ticks_number_bounds[0]
            )
            self.expected_ticks_number_upper_bound = (
                expected_ticks_number_bounds[1]
            )

    def _expected_number_of_ticks(self) -> float:
        """
        Calculate E[T] using an EWMA of previous bar tick counts.
        """
        prev_ticks_list = self.run_bars_statistics[PREVIOUS_BARS_N_TICKS_LIST]
        
        window = self.window_size_for_expected_n_ticks_estimation
        if window is None or window <= 0:
             return np.mean(prev_ticks_list)
             
        if not prev_ticks_list:
             return self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER]

        ewma_ticks = ewma(
            np.array(prev_ticks_list[-window:], dtype=float), window=window
        )[-1]

        return min(
            max(ewma_ticks, self.expected_ticks_number_lower_bound),
            self.expected_ticks_number_upper_bound,
        )


class FixedRunBars(AbstractRunBars):
    """
    Concrete class for Run Bars with a fixed (constant)
    Expected Number of Ticks (E[T]).
    """

    def __init__(
        self,
        bar_type: str,
        initial_estimate_of_expected_n_ticks_in_bar: int,
        window_size_for_expected_imbalance_estimation: int,
        analyse_thresholds: bool = False,
        window_size_for_expected_n_ticks_estimation: Optional[int] = None, # Not used
    ):
        """
        Constructor.

        Parameters
        ----------
        bar_type : str
            e.g., 'dollar_run', 'tick_run'.
        initial_estimate_of_expected_n_ticks_in_bar : int
            The *fixed* value for E[T].
        window_size_for_expected_imbalance_estimation : int
            Window size for EWMA of E[theta].
        analyse_thresholds : bool, default=False
            If True, store threshold data for analysis.
        window_size_for_expected_n_ticks_estimation : int, optional
            Ignored. Kept for consistent interface with factory.
        """
        super().__init__(
            bar_type,
            window_size_for_expected_n_ticks_estimation=None,
            window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
            initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
            analyse_thresholds=analyse_thresholds,
        )

    def _expected_number_of_ticks(self) -> float:
        """
        Return the fixed E[T] value.
        """
        return self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER]


"""
Implements the Probabilistic Sharpe Ratio (PSR) and related metrics
as described by Marcos Lopez de Prado.
"""

from typing import List
import numpy as np
from scipy import stats as ss

def probabilistic_sharpe_ratio(
    observed_sharpe_ratio: float,
    benchmark_sharpe_ratio: float,
    number_of_returns: int,
    skewness_of_returns: float = 0.0,
    kurtosis_of_returns: float = 3.0,
    return_test_statistic: bool = False,
) -> float:
    r"""
    Calculate the Probabilistic Sharpe Ratio (PSR).

    The PSR estimates the probability that the observed Sharpe ratio (SR)
    is greater than a benchmark SR (e.g., the expected maximum SR from
    multiple trials) given the track record length and return moments.

    The test statistic is:
    .. math::
        Z = \frac{(\hat{SR} - SR^*) \sqrt{T-1}}
                 {\sqrt{1 - S \hat{SR} + \frac{K-1}{4} \hat{SR}^2}}

    PSR = N(Z)

    Where:
    - \(\hat{SR}\) is the observed Sharpe ratio
    - \(SR^*\) is the benchmark Sharpe ratio
    - \(T\) is the number of returns
    - \(S\) is the skewness of returns
    - \(K\) is the kurtosis of returns

    Parameters
    ----------
    observed_sharpe_ratio : float
        The observed Sharpe ratio of the strategy.
    benchmark_sharpe_ratio : float
        The benchmark Sharpe ratio (e.g., E[max(SR)]).
    number_of_returns : int
        The number of return observations (T).
    skewness_of_returns : float, default=0.0
        The skewness of the returns (S).
    kurtosis_of_returns : float, default=3.0
        The kurtosis of the returns (K). Assumes 3 for normal.
    return_test_statistic : bool, default=False
        If True, return the Z-statistic instead of the CDF value (PSR).

    Returns
    -------
    float
        The Probabilistic Sharpe Ratio (if `return_test_statistic` is False)
        or the Z test statistic (if True).

    Example
    -------
    >>> psr = probabilistic_sharpe_ratio(
    ...     observed_sharpe_ratio=2.5,
    ...     benchmark_sharpe_ratio=1.0,
    ...     number_of_returns=252,
    ...     skewness_of_returns=-0.5,
    ...     kurtosis_of_returns=4.0
    ... )
    >>> print(f"PSR: {psr:.2f}")
    PSR: 1.00
    """
    denominator = (
        1
        - skewness_of_returns * observed_sharpe_ratio
        + (kurtosis_of_returns - 1) / 4 * observed_sharpe_ratio**2
    )

    # Handle cases where denominator is non-positive due to extreme inputs
    if denominator <= 0:
        return 0.0 if not return_test_statistic else -np.inf

    test_statistic = (
        (observed_sharpe_ratio - benchmark_sharpe_ratio)
        * np.sqrt(number_of_returns - 1)
    ) / np.sqrt(denominator)

    if return_test_statistic:
        return test_statistic

    return ss.norm.cdf(test_statistic)


def benchmark_sharpe_ratio(sharpe_ratio_estimates: List[float]) -> float:
    r"""
    Calculate the expected maximum Sharpe Ratio (Benchmark SR).

    This is used as the benchmark SR* in the PSR calculation. It represents
    the expected maximum SR one would observe from N independent trials.

    .. math::
        SR^* = \sigma_{SR} \left[ (1 - \gamma) \Phi^{-1}(1 - \frac{1}{N})
               + \gamma \Phi^{-1}(1 - \frac{1}{N} e^{-1}) \right]

    Where:
    - \(\sigma_{SR}\) is the standard deviation of SR estimates
    - \(\gamma\) is the Euler-Mascheroni constant
    - \(\Phi^{-1}\) is the inverse CDF of a standard normal distribution
    - \(N\) is the number of SR estimates (trials)

    Parameters
    ----------
    sharpe_ratio_estimates : List[float]
        A list or array of Sharpe ratio estimates from N different trials.

    Returns
    -------
    float
        The Benchmark Sharpe Ratio (E[max(SR)]).

    Example
    -------
    >>> sr_list = [0.5, 1.2, -0.3, 0.8, 1.5, 0.9]
    >>> bsr = benchmark_sharpe_ratio(sr_list)
    >>> print(f"Benchmark SR: {bsr:.2f}")
    Benchmark SR: 1.03
    """
    n_estimates = len(sharpe_ratio_estimates)
    if n_estimates <= 1:
        return np.mean(sharpe_ratio_estimates) if n_estimates == 1 else 0.0

    standard_deviation = np.array(sharpe_ratio_estimates).std()

    term1 = (1 - np.euler_gamma) * ss.norm.ppf(1 - 1 / n_estimates)
    term2 = np.euler_gamma * ss.norm.ppf(1 - 1 / (n_estimates * np.e))

    benchmark_value = standard_deviation * (term1 + term2)

    return benchmark_value


"""
High-Performance Computing (HPC) utilities for parallel processing.

Provides a set of functions to parallelize operations, especially
on pandas objects, using Python's multiprocessing.
"""

import multiprocessing as mp
import pandas as pd
import numpy as np
import datetime as dt
import time
from typing import List, Dict, Any, Callable, Tuple, Union

def report_progress(
    job_number: int, total_jobs: int, start_time: float, task: str
) -> None:
    """
    Report the progress of a computing task to the console.

    Parameters
    ----------
    job_number : int
        The current job number (e.g., `i` in a loop).
    total_jobs : int
        The total number of jobs.
    start_time : float
        The time the process started (e.g., `time.time()`).
    task : str
        A description of the task being performed.
    """
    progress = job_number / total_jobs
    elapsed_time_min = (time.time() - start_time) / 60.0
    remaining_time_min = elapsed_time_min * (1 / progress - 1)
    
    timestamp = str(dt.datetime.fromtimestamp(time.time()))
    message = (
        f"{timestamp} {progress*100:.2f}% {task} done after "
        f"{elapsed_time_min:.2f} minutes. Remaining {remaining_time_min:.2f} minutes."
    )
    
    if job_number < total_jobs:
        print(message, end='\r')
    else:
        print(message)


def expand_call(kargs: Dict[str, Any]) -> Any:
    """
    Wrapper function to expand keyword arguments for a callback.

    This is used by `process_jobs` to unpack a dictionary of arguments
    and call the target function.

    Parameters
    ----------
    kargs : dict
        A dictionary of arguments, which *must* include a 'func' key
        mapping to the function to be called.

    Returns
    -------
    Any
        The output of the callback function.
    """
    func = kargs.pop('func')
    return func(**kargs)


def process_jobs(
    jobs: List[Dict[str, Any]], task: Optional[str] = None, num_threads: int = 24
) -> List[Any]:
    """
    Process a list of jobs in parallel using multiprocessing.

    Parameters
    ----------
    jobs : List[Dict[str, Any]]
        A list of job dictionaries. Each dict must contain a 'func' key
        and all arguments required by that function.
    task : str, optional
        A name for the task, used for progress reporting. If None,
        the function name from the first job is used.
    num_threads : int, default=24
        The number of parallel processes to spawn.

    Returns
    -------
    List[Any]
        A list containing the results from all jobs.
    """
    if task is None:
        task = jobs[0]['func'].__name__

    with mp.Pool(processes=num_threads) as pool:
        outputs = []
        start_time = time.time()
        
        # Use imap_unordered for efficient processing
        imap_results = pool.imap_unordered(expand_call, jobs)
        
        # Process results as they complete
        for i, result in enumerate(imap_results, 1):
            outputs.append(result)
            report_progress(i, len(jobs), start_time, task)
            
    return outputs


def process_jobs_sequential(jobs: List[Dict[str, Any]]) -> List[Any]:
    """
    Run jobs sequentially (single-thread). Useful for debugging.

    Parameters
    ----------
    jobs : List[Dict[str, Any]]
        A list of job dictionaries.

    Returns
    -------
    List[Any]
        A list containing the results from all jobs.
    """
    output = []
    for job in jobs:
        output_ = expand_call(job)
        output.append(output_)
    return output


def linear_partitions(num_atoms: int, num_threads: int) -> np.ndarray:
    """
    Generate linear partitions (indices) for parallel computation.

    Splits `num_atoms` into `num_threads` (or fewer) roughly equal parts.

    Parameters
    ----------
    num_atoms : int
        The total number of items to split.
    num_threads : int
        The desired number of partitions (threads).

    Returns
    -------
    np.ndarray
        An array of partition boundary indices.
    """
    n_parts = min(num_threads, num_atoms)
    partitions = np.linspace(0, num_atoms, n_parts + 1)
    partitions = np.ceil(partitions).astype(int)
    return partitions


def nested_partitions(
    num_atoms: int, num_threads: int, upper_triangle: bool = False
) -> np.ndarray:
    r"""
    Generate nested partitions, useful for tasks with nested loops.

    The formula is designed to create partitions of varying sizes,
    which can be more efficient for tasks where work is not
    distributed linearly (e.g., matrix operations).

    .. math::
       p_i = \frac{-1 + \sqrt{1 + 4(p_{i-1}^2 + p_{i-1} + N(N+1)/K)}}{2}

    Parameters
    ----------
    num_atoms : int
        Number of atoms (N).
    num_threads : int
        Number of threads (K).
    upper_triangle : bool, default=False
        If True, sort partitions to be heaviest first.

    Returns
    -------
    np.ndarray
        An array of partition boundary indices.
    """
    partitions = [0]
    n_threads_ = min(num_threads, num_atoms)
    
    for _ in range(n_threads_):
        last_part = partitions[-1]
        part_size = 1 + 4 * (
            last_part**2 + last_part + num_atoms * (num_atoms + 1.0) / n_threads_
        )
        part_val = (-1 + part_size**0.5) / 2.0
        partitions.append(part_val)
        
    partitions = np.round(partitions).astype(int)
    
    if upper_triangle:  # The first rows are the heaviest
        partitions = np.cumsum(np.diff(partitions)[::-1])
        partitions = np.append(np.array([0]), partitions)
        
    return partitions


def mp_pandas_obj(
    func: Callable[..., pd.Series],
    pandas_object: Tuple[str, pd.Index],
    num_threads: int = 2,
    mp_batches: int = 1,
    linear_partition: bool = True,
    **kwargs: Any
) -> Union[pd.DataFrame, pd.Series]:
    """
    Parallelize a function call on a pandas object (DataFrame/Series).

    This function splits the `pandas_object` index into partitions and
    calls `func` on each partition in parallel.

    Parameters
    ----------
    func : Callable
        The function to be parallelized. It must accept a pandas
        object as its first argument.
    pandas_object : Tuple[str, pd.Index]
        A tuple where:
        - [0] (str): The name of the argument in `func` that
                     receives the partition (e.g., 'molecule').
        - [1] (pd.Index): The index to be partitioned.
    num_threads : int, default=2
        Number of parallel processes.
    mp_batches : int, default=1
        Number of batches to split the jobs into.
    linear_partition : bool, default=True
        If True, use `linear_partitions`.
        If False, use `nested_partitions`.
    **kwargs : Any
        Other keyword arguments to be passed to `func`.

    Returns
    -------
    Union[pd.DataFrame, pd.Series]
        The concatenated results from all parallel calls.
    """
    if linear_partition:
        parts = linear_partitions(
            len(pandas_object[1]), num_threads * mp_batches
        )
    else:
        parts = nested_partitions(
            len(pandas_object[1]), num_threads * mp_batches
        )

    jobs = []
    for i in range(1, len(parts)):
        job = {
            pandas_object[0]: pandas_object[1][parts[i - 1] : parts[i]],
            "func": func,
        }
        job.update(kwargs)
        jobs.append(job)

    if num_threads == 1:
        out = process_jobs_sequential(jobs)
    else:
        out = process_jobs(jobs, num_threads=num_threads)

    if isinstance(out[0], pd.DataFrame):
        result_df = pd.concat(out)
    elif isinstance(out[0], pd.Series):
        result_df = pd.concat(out)
    else:
        return out

    return result_df.sort_index()


"""
Computes Mean Decrease Impurity (MDI) feature importance.
"""

import pandas as pd
import numpy as np
from typing import List, Optional, Union, Any
from sklearn.ensemble import BaseEnsemble
from .feature_importance_strategy import FeatureImportanceStrategy

class FeatureImportanceMDI(FeatureImportanceStrategy):
    """
    Computes feature importance using Mean Decrease Impurity (MDI).

    This method is specific to tree-based ensembles (like RandomForest)
    and measures importance as the average impurity decrease.
    """

    def __init__(self, classifier: BaseEnsemble):
        """
        Initialize the strategy.

        Parameters
        ----------
        classifier : BaseEnsemble
            An *untrained* scikit-learn ensemble model (e.g., RandomForestClassifier).
        """
        if not hasattr(classifier, 'estimators_'):
            raise TypeError("Classifier must be an ensemble (e.g., RandomForest).")
        self.classifier = classifier

    def compute(self, x: pd.DataFrame, y: pd.Series, **kwargs: Any) -> pd.DataFrame:
        """
        Compute MDI feature importance.

        Parameters
        ----------
        x : pd.DataFrame
            The feature data.
        y : pd.Series
            The target data.
        **kwargs : Any
            Keyword arguments for the classifier's `fit` method
            (e.g., `sample_weight`).

        Returns
        -------
        pd.DataFrame
            DataFrame with "Mean" and "StandardDeviation" of importance.
        """
        train_sample_weights = kwargs.get('sample_weight')
        
        # Fit the classifier
        self.classifier.fit(x, y, sample_weight=train_sample_weights)

        # Get importance from each tree
        importances_dict = {
            i: tree.feature_importances_
            for i, tree in enumerate(self.classifier.estimators_)
        }
        importances_df = pd.DataFrame.from_dict(importances_dict, orient="index")
        
        # Ensure correct feature names
        if hasattr(self.classifier, 'feature_names_in_'):
             importances_df.columns = self.classifier.feature_names_in_
        else:
             importances_df.columns = x.columns

        # Replace 0 with NaN (as per user's original code)
        importances_df.replace(0, np.nan, inplace=True)

        importances = pd.concat(
            {
                "Mean": importances_df.mean(),
                "StandardDeviation": (
                    importances_df.std()
                    * (importances_df.shape[0] ** -0.5)
                ),
            },
            axis=1,
        )

        # Normalize
        importances /= importances["Mean"].sum()
        return importances


"""
RiskLabAI Synthetic Data Module

Provides tools for generating synthetic financial data, including:
- Drift-Burst Hypothesis (DBH) model.
- Heston-Merton model with Markov-switching regimes.
"""

from .drift_burst_hypothesis import drift_volatility_burst
from .synthetic_controlled_environment import (
    compute_log_returns,
    heston_merton_log_returns,
    align_params_length,
    generate_prices_from_regimes,
    parallel_generate_prices,
)

__all__ = [
    "drift_volatility_burst",
    "compute_log_returns",
    "heston_merton_log_returns",
    "align_params_length",
    "generate_prices_from_regimes",
    "parallel_generate_prices",
]


"""
Implements Time Bars.
"""

from typing import Union, List, Any, Iterable
import numpy as np
import pandas as pd
from RiskLabAI.data.structures.abstract_bars import AbstractBars, TickData

class TimeBars(AbstractBars):
    """
    Concrete class for Time Bars.

    Generates a new bar whenever a fixed time duration has passed.
    """

    def __init__(self, resolution_type: str, resolution_units: int):
        """
        TimeBars constructor.

        Parameters
        ----------
        resolution_type : str
            Type of bar resolution: 'S', 'MIN', 'H', 'D', 'W'.
        resolution_units : int
            Number of units for the resolution (e.g., 5 for 5-minutes).
        """
        super().__init__(bar_type="time")

        self.resolution_to_n_seconds = {
            "S": 1,
            "MIN": 60,
            "H": 3600,
            "D": 86400,
            "W": 604800,
        }

        if resolution_type.upper() not in self.resolution_to_n_seconds:
            raise ValueError(f"Invalid resolution_type. Use one of {list(self.resolution_to_n_seconds.keys())}")
            
        self.resolution_type = resolution_type.upper()
        self.resolution_units = resolution_units
        self.threshold_in_seconds = (
            self.resolution_units * self.resolution_to_n_seconds[self.resolution_type]
        )
        
        self.current_bar_timestamp = np.nan
        self.current_bar_end_timestamp = np.nan

    def construct_bars_from_data(self, data: Iterable[TickData]) -> List[List[Any]]:
        """
        Constructs time bars from input tick data.

        Parameters
        ----------
        data : Iterable[TickData]
            An iterable (list, tuple, generator) of tick data.
            Each tick is (date_time, price, volume). 
            `date_time` must be a pandas Timestamp.

        Returns
        -------
        List[List[Any]]
            A list of the constructed time bars.
        """
        bars_list = []
        for tick_data in data:
            self.tick_counter += 1

            # Unpack data
            date_time, price, volume = tick_data[0], tick_data[1], tick_data[2]
            
            # Get tick timestamp in seconds
            tick_timestamp_sec = date_time.timestamp()
            
            # Determine the "floor" timestamp for this bar
            bar_start_timestamp_sec = (
                int(tick_timestamp_sec // self.threshold_in_seconds)
                * self.threshold_in_seconds
            )
            
            # Initialize first bar
            if np.isnan(self.current_bar_timestamp):
                self.current_bar_timestamp = bar_start_timestamp_sec
                self.current_bar_end_timestamp = (
                    bar_start_timestamp_sec + self.threshold_in_seconds
                )
            
            # Check if this tick belongs to a new bar
            if self._bar_construction_condition(tick_timestamp_sec):
                # Construct the *previous* bar
                bar_end_time = pd.to_datetime(self.current_bar_end_timestamp, unit='s')
                
                next_bar = self._construct_next_bar(
                    bar_end_time,
                    self.tick_counter - 1, # Index of the *previous* tick
                    self.close_price,      # Close price from *previous* tick
                    self.high_price,
                    self.low_price,
                    self.current_bar_end_timestamp,
                )
                bars_list.append(next_bar)

                # Reset for the new bar
                self._reset_cached_fields()
                self.current_bar_timestamp = bar_start_timestamp_sec
                self.current_bar_end_timestamp = (
                    bar_start_timestamp_sec + self.threshold_in_seconds
                )

            # Update fields with current tick data
            tick_rule = self._tick_rule(price)
            self.update_base_fields(price, tick_rule, volume)
            self.close_price = price

        return bars_list

    def _bar_construction_condition(self, tick_timestamp_sec: float) -> bool:
        """
        Check if the current tick's timestamp has crossed
        the end-time of the current bar.
        """
        return tick_timestamp_sec >= self.current_bar_end_timestamp


"""
Implements the Lempel-Ziv (LZ) Entropy estimator.
"""

def lempel_ziv_entropy(message: str) -> float:
    """
    Calculate the Lempel-Ziv (LZ) complexity as an entropy estimator.

    This function calculates the number of unique substrings encountered
    during a one-pass traversal of the message, normalized by the
    message length.

    Parameters
    ----------
    message : str
        Input string.

    Returns
    -------
    float
        The calculated Lempel-Ziv Entropy.
    """
    if not message:
        return 0.0

    library = set()
    message_length = len(message)
    i = 0

    while i < message_length:
        j = i
        # Find the longest substring starting at `i` that is *not* in the library
        while j < message_length and message[i : j + 1] in library:
            j += 1
        
        # Add the new, unseen substring to the library
        library.add(message[i : j + 1])
        i = j + 1

    return len(library) / message_length


"""
RiskLabAI Optimization Module

Implements advanced portfolio optimization techniques, including
Nested Clustered Optimization (NCO) and Hierarchical Risk Parity (HRP).
"""

from .hrp import (
    inverse_variance_weights,
    cluster_variance,
    quasi_diagonal,
    recursive_bisection,
    distance_corr,
    hrp,
)
from .nco import (
    get_optimal_portfolio_weights,
    get_optimal_portfolio_weights_nco,
)
from .hyper_parameter_tuning import (
    MyPipeline,
    clf_hyper_fit,
)

__all__ = [
    # HRP
    "inverse_variance_weights",
    "cluster_variance",
    "quasi_diagonal",
    "recursive_bisection",
    "distance_corr",
    "hrp",
    
    # NCO
    "get_optimal_portfolio_weights",
    "get_optimal_portfolio_weights_nco",
    
    # Tuning
    "MyPipeline",
    "clf_hyper_fit",
]